/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS Packet buffer API
 *
 * A packet buffer (pbuf) is a NodeOS abstraction representing a contiguous
 * memory buffer.  Pbufs are much like Linux skbuffs in that they may be
 * allocated with extra space both before and after the "active" part of
 * the buffer, to allow for expansion at either end (e.g. for prepending
 * headers).  This extra space is tracked by keeping an offset and size
 * of the active area of the packet:
 *
 * "head" room: an_pbuf_offset(pb)
 * "tail" room: an_pbuf_bufsize(pb) - (an_pbuf_offset(pb) + an_pbuf_size(pb))
 *
 * This space may be directly accessed by the application, but the application
 * must "sync" its changes to the offset or size with the NodeOS (using
 * an_pbuf_setoffset(pb, off) and an_pbuf_setsize(pb, size) before performing
 * any NodeOS calls involving the buffer.
 *
 * Note that though an EE may specify its own values for this space when
 * creating buffers, those values are only hints to the NodeOS.  The NodeOS
 * may ignore them and choose its own values (including no space).
 */

#ifndef _NODEOS_PBUF_H_
#define _NODEOS_PBUF_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_Pbuf an_pbuf_t
#define an_pbufCreate an_pbuf_create
#define an_pbufDestroy an_pbuf_destroy
#define an_pbufData an_pbuf_data
#define an_pbufSize an_pbuf_size
#define an_pbufOffset an_pbuf_offset
#define an_pbufBufSize an_pbuf_bufsize
#define an_pbufSetOffset an_pbuf_setoffset
#define an_pbufSetSize an_pbuf_setsize
#endif

#include <nodeos/an_error.h>
#include <nodeos/impl/an_pbuf.h>

/* create/init flags */
#define AN_PBUF_SIZEMASK	3
#define AN_PBUF_DEFAULT		0	/* suitable default value */
#define AN_PBUF_ETH		1	/* for ethernet packets */
#define AN_PBUF_SOCK		2	/* for UDP packets */
#define AN_PBUF_CLOCK		3	/* for "clock" packets */

/*
 * Initializes the provided memory as a pbuffer object.
 * Characteristics of the buffer (e.g., its size) are described
 * by the flags parameter.  The offset and size parameters provide hints
 * to the NodeOS about how much room to leave before and after the
 * expected "active" area of the buffer.  By setting the offset non-zero,
 * the buffer may later be expanded at the head by an_pbuf_setoffset.
 * Likewise, if size is less than the (fixed) size of the buffer,
 * the buffer may be expanded at the tail by using an_pbuf_setsize. 
 *
 * Memory for the buffer is allocated by the NodeOS and charged
 * to the current flow.  Returns zero if successful, an error otherwise.
 */
an_pbuf_t *an_pbuf_create(an_pbuf_t *mem, int flags, int offset, int size);

/*
 * Destroys a packet buffer and returns resources to the NodeOS.
 * After deinit, the pbuf object is no longer valid and the memory it
 * occupied can be reused; i.e., memory for the object itself is not freed,
 * only the actual buffer memory.  The effect of destroying a pbuf which
 * is in use is undefined but not likely to be good.
 */
an_error_t an_pbuf_destroy(an_pbuf_t *pbuf);

/*
 * Returns a pointer to the beginning of the currently valid data
 * region of the given buffer.
 */
void *an_pbuf_data(an_pbuf_t *pbuf);

/*
 * Returns the size of the currently valid data region of the given buffer.
 */
int an_pbuf_size(an_pbuf_t *pbuf);

/*
 * Returns the offset in the buffer of the active data region.
 */
int an_pbuf_offset(an_pbuf_t *pbuf);

/*
 * Returns the actual size of the buffer.
 */
int an_pbuf_bufsize(an_pbuf_t *pbuf);

/*
 * Change the offset of the active data region to the indicated value.
 * Implicitly changes the size of the active data region.  If successful,
 * this call returns a pointer to the new region in ndata and the call
 * returns zero.  Otherwise, a non-zero error code is returned.
 */
an_error_t an_pbuf_setoffset(an_pbuf_t *pbuf, int noffset, void **ndata);

/*
 * Change the size of the active data region to the indicated value.
 * If successful, this call returns a pointer to the new region in ndata
 * and the call returns zero.  Otherwise, a non-zero error code is
 * returned.
 */
an_error_t an_pbuf_setsize(an_pbuf_t *pbuf, int nsize, void **ndata);

#endif /* _NODEOS_PBUF_H_ */
