/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Implementation specific types and constants for the Utah OSKit-based NodeOS.
 *
 * Note that these are not machine dependencies, this file is used for both
 * the x86 and ARM versions.
 *
 * Note that this file also currently contains "OS" specific constants and
 * types; i.e., for Linux, FreeBSD and Solaris as well as for OSKit.
 */
#ifndef _NODEOS_IMPL_IMPL_H_
#define _NODEOS_IMPL_IMPL_H_

#ifdef NODEOS_SPEC_COMPAT
#define an_NotifyFunc an_flowterm_func_t
#define an_FlowNotifyFunc an_flowterm_func_t
#define an_FlowStartSpec an_flowinit_t
#define an_MempoolNotifyFunc an_mempoolfull_func_t
#define an_ChannelAddress an_chanaddr_t
#define an_ChannelProtocol an_chanproto_t
#define an_DeliverFunc an_chanrecvfunc_t
#define an_Schedule an_sched_t
#define an_ThreadNotifyFunc an_threadintr_func_t
#define an_threadSetHandler an_thread_sethandler
#endif

#include <sys/types.h>
#include <sys/socket.h>

/*
 * XXX hack global constants that should be specified in other ways.
 */
#define NODEOS_PAGESIZE		4096	/* memory page size */
#define NODEOS_CHAN_PROTOSIZE	128	/* protocol spec size */
#define NODEOS_CHAN_ADDRSIZE	128	/* peer addressing info size */
#define NODEOS_MAXIFS		16	/* maximum network interfaces */
#define NODEOS_MAXIFNAME	8	/* maximum interface pname length */

#define NODEOS_CHAN_ETH_BSIZE	2048		/* eth channel buffer size */
#define NODEOS_CHAN_ETH_CSIZE	128		/* size under which copied */
#define NODEOS_CHAN_CLK_BSIZE	4		/* clock channel buffer size */
#define NODEOS_CHAN_SOCK_BSIZE	(64*1024)	/* sock channel buffer size */

/*
 * Implementation-specific, API-exposed types.
 */

/*
 * An ethernet-style channel address
 */
typedef struct {
	u_char daddr[6];
	u_char saddr[6];
	u_short type;
} an_ethaddr_t;

/*
 * A BSD-socket-style channel address
 */
typedef struct sockaddr an_sockaddr_t;

/*
 * A cheezy-hack-clock-style channel address
 */
typedef struct {
	int freq;
} an_clockaddr_t;

/*
 * Implementation-specific, API-exposed types.
 */

/*
 * Flow initialization and termination functions.
 */
struct an_flow;
typedef void (*an_flowinit_func_t)(void *);
typedef void (*an_flowterm_func_t)(struct an_flow *);

/*
 * Called when a mempool is over limit
 */
struct an_mempool;
typedef void (*an_mempoolfull_func_t)(struct an_mempool *, int, int);

/*
 * Channel protocol processing information.
 * Currently implemented as a string; e.g. "ETH/IP".
 */
typedef struct {
	int len;
	char data[NODEOS_CHAN_PROTOSIZE-sizeof(int)];
} an_chanproto_t;

/*
 * Channel peer address information.
 * Channel-type specific.
 */
typedef union {
	an_ethaddr_t eth;
	an_sockaddr_t socket;
	an_clockaddr_t clock;
	char data[NODEOS_CHAN_ADDRSIZE];
} an_chanaddr_t;

/*
 * Channel statistics
 */
typedef struct {
	int inpackets;		/* calls to recv function */
	unsigned int inbytes;	/* and number of bytes received */
	int outpackets;		/* calls to ani_chan_send */
	unsigned int outbytes;	/* and number of bytes sent */
	int droppackets;	/* total packets dropped by NodeOS.. */
	int dropnobufs;		/* ..due to no memory resources */
	int dropfailedthread;	/* ..due to failed thread start */
	int dropfailedbuf;	/* ..due to failed buffer setup */
	int bufqueued;		/* packets queued due to no threads */
	int bufcopies;		/* inpackets needing extra copy */
	int maxthreads;		/* max threads active */
} an_chanstats_t;

/*
 * Called when a packet arrives on a channel.
 * The "void *" is an additional argument supplied by the user
 * at channel creation time.
 */
struct an_pbuf;
typedef void (*an_chanrecvfunc_t)(struct an_pbuf *, void *);

/*
 * Not yet
 */
typedef void *an_sched_t;

/*
 * Thread interrupt handling
 * Includes API call to set the handler and signature for handler.
 */
typedef void (*an_threadintr_func_t)(int, void *);

/* types passed to the notify function */
#define AN_THREAD_TRAP_SEGV 0x11
#define AN_THREAD_TRAP_FPU  0x12  
#define AN_THREAD_INTR	    0x21

void an_thread_sethandler(an_threadintr_func_t handler);

#endif /* _NODEOS_IMPL_IMPL_H_ */
