/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * NodeOS internal representation of a channel.
 */

#ifndef _CHANNEL_IMPL_H_
#define _CHANNEL_IMPL_H_

#define CHANNEL_STATS

#include <nodeos/an_impl.h>
#include <nodeos/an_resource.h>
#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_pbuf.h>

struct chanops;
struct ani_demuxkey;
struct ani_flow;
struct ani_mempool;
struct threadpool;

struct ani_chan {
	an_chanaddr_t	 addr;	   /* peer addressing info */
	an_chanproto_t	 proto;	   /* protocol processing info */
	struct ani_chan	 *next;	   /* link of all channels for a flow */
	struct ani_flow	 *flow;	   /* flow we are associated with */
	struct threadpool *tpool;  /* thread pool for our CPU resources */
	int		 curthreads;/* threads processing packets */
	int		 maxthreads;/* max threads allowed */
	struct chanops	 *ops;	   /* operations */
	struct ani_mempool *mpool; /* memory pool for our memory resources */
	int		 flags;	   /* channel type and assorted flags */
	union {
		struct {
			struct ani_demuxkey *demux;/* demux information */
			ani_pbuf_t	*rbufs;	   /* free pbufs */
			ani_pbuf_t	*qbufs;	   /* queued, full pbufs */
			an_chanrecvfunc_t func;	   /* delivery function */
			void		*farg;	   /* delivery function arg */
		} in;
		struct {
			struct ani_chan	*inchan;   /* component in-chan */
			struct ani_chan	*outchan;  /* component out-chan */
		} cut;
	}		 tdata;	   /* channel type specific data */
	void		 *data;	   /* channel-protocol specific data */
	int		 refs;
	ani_ilock_t	 lock;
	an_chanstats_t	 stats;	   /* channel statistics */	
};
typedef struct ani_chan ani_chan_t;

/* channel types */
#define CHAN_IN		1
#define CHAN_OUT	2
#define CHAN_CUT	3
#define CHAN_TMASK	3

/* channel flags */
#define CHAN_DYING	0x04	/* channel has been destroyed */
#define CHAN_WITHCUT	0x08	/* associated with a cut-through channel */
#define CHAN_ISRAW	0x10	/* provide protocol headers (if possible) */

#define ani_chan_lock(c)		ani_ilock_lock(&(c)->lock)
#define ani_chan_trylock(c)	ani_ilock_trylock(&(c)->lock)
#define ani_chan_unlock(c)	ani_ilock_unlock(&(c)->lock)
#define ani_chan_islocked(c)	ani_ilock_islocked(&(c)->lock)

#define ani_chan_maxthreads(c)	((c)->maxthreads)
#define ani_chan_name(c)	((char *)((c)->proto.data))
#define ani_chan_namelen(c)	((c)->proto.len)
#define ani_chan_type(c)	((c)->flags & CHAN_TMASK)
#define ani_chan_mempool(c)	((c)->mpool)

#define ani_chan_reference(c)	((c)->refs++)
#define ani_chan_release(c)	if (--(c)->refs == 0) ani_chan_cleanup(c)

#ifdef MEMDEBUG
#define assert_ani_chan_valid(c) \
	assert((c) != 0 && (c) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(c) != 0xdddddddd)
#else
#define assert_ani_chan_valid(c)	assert((c) != 0)
#endif

#ifdef CHANNEL_STATS
#define CHANSTATS(s)	s
#else
#define CHANSTATS(s)
#endif

struct chanops {
	const char *name;
	void (*init)(void);
	int (*create)(ani_chan_t *);
	void (*destroy)(ani_chan_t *);
	void (*send)(ani_chan_t *, ani_pbuf_t *);
	int (*enable)(ani_chan_t *);
	void (*disable)(ani_chan_t *);
};

struct demux;
struct an_pbuf;

void ani_chan_setup(void);
int ani_chan_init(ani_chan_t *chan, int ctype, struct ani_flow *flow,
		  an_chanproto_t *proto, an_chanaddr_t *addr,
		  struct ani_demuxkey *demux, an_cpuspec_t *cpu,
		  ani_pbuf_t **pbufs, int npbufs, an_netspec_t *net,
		  an_chanrecvfunc_t func, void *farg);
int ani_chan_deinit(ani_chan_t *chan);
void ani_chan_cleanup(ani_chan_t *chan);
int ani_chan_enable(ani_chan_t *chan);
void ani_chan_disable(ani_chan_t *chan);
int ani_chan_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
an_chanstats_t *ani_chan_stats(ani_chan_t *chan);
int ani_chan_thread_start(ani_chan_t *chan, ani_pbuf_t *pbuf);
void ani_chan_thread_exit(ani_chan_t *chan);
int ani_chan_pbuf_alloc(ani_chan_t *chan, size_t size, ani_pbuf_t **out_pbuf);
void ani_chan_pbuf_free(ani_chan_t *chan, ani_pbuf_t *pbuf);
int ani_chan_pbuf_add(ani_chan_t *chan, ani_pbuf_t *pbufs[], int npbufs);
void ani_chan_pbuf_removeall(ani_chan_t *chan);
int ani_chan_demux_add(ani_chan_t *chan, struct demux **listp);
void ani_chan_demux_remove(ani_chan_t *chan, struct demux **listp);

int ani_cutchan_init(ani_chan_t *chan, struct ani_flow *flow,
		     an_chanproto_t *proto,
		     ani_chan_t *inchan, ani_chan_t *outchan,
		     an_chanrecvfunc_t rfunc);
void ani_cutchan_receive(struct an_pbuf *pbuf, void *arg);
#ifdef CHANNEL_STATS
void ani_chan_stats_dump(ani_chan_t *chan);
#endif

#endif /* _CHANNEL_IMPL_H_ */
