/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The packet buffer (pbuf) API
 */
#include <nodeos/an_pbuf.h>
#include <nodeos/an_error.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_mempool.h>

#include "os.h"

an_pbuf_t *
an_pbuf_create(an_pbuf_t *mem, int flags, int offset, int size)
{
	ani_flow_t *iflow = ani_flow_current();
	ani_pbuf_t *ipbuf = 0;
	size_t bsize;
	ani_mempool_t *cpool;
	int err, didalloc = 0;

	NODEOS_ENTER;

	/*
	 * Translate size-related flags.
	 *
	 * XXX stinky hack.
	 */
	switch (flags & AN_PBUF_SIZEMASK) {
	case AN_PBUF_DEFAULT:
		bsize = NODEOS_CHAN_ETH_BSIZE;
		break;
	case AN_PBUF_CLOCK:
		bsize = NODEOS_CHAN_CLK_BSIZE;
		break;
	case AN_PBUF_ETH:
		bsize = NODEOS_CHAN_ETH_BSIZE;
		break;
	case AN_PBUF_SOCK:
		bsize = NODEOS_CHAN_SOCK_BSIZE;
		break;
	default:
		panic("pbuf_init: unknown size");
		bsize = 0;
	}

	/*
	 * Constrain the active region
	 */
	if (offset < 0 || size < 0 || offset + size > bsize) {
		err = AN_ERR_INVALIDARG;
		goto done;
	}

	/*
	 * Allocate memory for the object if needed.
	 */
	cpool = ani_flow_mempool(iflow);
	if (mem == 0) {
		mem = ani_mempool_alloc(cpool, sizeof(an_pbuf_t));
		didalloc = 1;
	}

	ipbuf = NODEOS_INTERN_PBUF(mem);
	if (ipbuf == 0)
		err = AN_ERR_INVALIDARG;
	else
		err = ani_pbuf_init(ipbuf, iflow, 0, bsize, offset, size);

	if (err != 0 && didalloc)
		ani_mempool_free(cpool, mem, sizeof(an_pbuf_t));

 done:
	NODEOS_LEAVE;
	return err ? 0 : NODEOS_EXTERN_PBUF(ipbuf);
}

an_error_t
an_pbuf_destroy(an_pbuf_t *pbuf)
{
	int err = 0;
	ani_pbuf_t *ipbuf;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	if (ipbuf == 0)
		err = AN_ERR_INVALIDARG;
	else
		ani_pbuf_deinit(ipbuf);
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_pbuf_setoffset(an_pbuf_t *pbuf, int noffset, void **ndata)
{
	void *data;
	ani_pbuf_t *ipbuf;
	int err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	err = ani_pbuf_setoffset(ipbuf, noffset, &data);
	if (err == 0 && ndata != 0)
		NODEOS_COPYOUT(ndata, &data, sizeof data, &err);
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_pbuf_setsize(an_pbuf_t *pbuf, int nsize, void **ndata)
{
	void *data;
	ani_pbuf_t *ipbuf;
	int err;

	NODEOS_ENTER;
	ipbuf = NODEOS_INTERN_PBUF(pbuf);
	err = ani_pbuf_setsize(ipbuf, nsize, &data);
	if (err == 0 && ndata != 0)
		NODEOS_COPYOUT(ndata, &data, sizeof data, &err);
	NODEOS_LEAVE;
	return NODEOS_EXTERN_ERROR(err);
}
