/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * The Thread API
 */
#include <nodeos/an_error.h>
#include <nodeos/an_thread.h>

#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_flow.h>
#include "threadpool.h"

an_error_t
an_thread_start(an_thread_t **out_thread, an_flow_t *flow,
	       void (*func)(void *arg), void *arg)
{
	ani_flow_t *iflow;
	ani_thread_t *ith;
	int err;

	NODEOS_ENTER;

	iflow = NODEOS_INTERN_FLOW(flow);
	if (iflow == 0 || out_thread == 0 || func == 0) {
		err = AN_ERR_INVALIDARG;
		goto done;
	}

	err = threadpool_thread_start(iflow->tpool, 0, (thread_func_t)func,
				      arg, 0, &ith);
	if (err == 0)
		*out_thread = NODEOS_EXTERN_THREAD(ith);

 done:
	NODEOS_LEAVE;
	return err;
}

void
an_thread_exit(void)
{
	NODEOS_ENTER;
	ani_thread_exit();
	NODEOS_LEAVE;
}

void
an_thread_yield(void)
{
	NODEOS_ENTER;
	ani_thread_yield();
	NODEOS_LEAVE;
}

void
an_thread_sleep(an_timespec_t *delay)
{
	NODEOS_ENTER;
	ani_thread_sleep(delay);
	NODEOS_LEAVE;
}

void
an_thread_setdata(void *data)
{
	NODEOS_ENTER;
	ani_thread_setdata(data);
	NODEOS_LEAVE;
}

void *
an_thread_getdata(void)
{
	void *data;

	NODEOS_ENTER;
	data = ani_thread_getdata();
	NODEOS_LEAVE;

	return data;
}

void
an_thread_intr(an_thread_t *thread)
{
	ani_thread_t *ithread;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread != 0)
		ani_thread_interrupt(ithread);
	NODEOS_LEAVE;
}

void
an_thread_blockintr(void)
{
	NODEOS_ENTER;
	ani_thread_block_interrupt();
	NODEOS_LEAVE;
}

void
an_thread_unblockintr(void)
{
	NODEOS_ENTER;
	ani_thread_unblock_interrupt();
	NODEOS_LEAVE;
}

an_error_t
an_thread_setprio(an_thread_t *thread, int prio)
{
	ani_thread_t *ithread;
	int err;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = AN_ERR_INVALIDARG;
	else
		err = ani_thread_setprio(ithread, prio);
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

an_error_t
an_thread_getprio(an_thread_t *thread, int *priop)
{
	ani_thread_t *ithread;
	int iprio, err;

	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = AN_ERR_INVALIDARG;
	else {
		iprio = ani_thread_getprio(ithread);
		NODEOS_COPYOUT(priop, &iprio, sizeof *priop, &err);
	}
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}

/*
 * Implementation specific API calls
 */

/*
 * Establish a thread_interrupt handler.
 */
void
an_thread_sethandler(an_threadintr_func_t handler)
{
	NODEOS_ENTER;
	ani_thread_sethandler(handler);
	NODEOS_LEAVE;
}

/*
 * Get various bits of stack information on a specific
 * thread.  The thread may be the current thread.
 * If the target thread is running (on an SMP machine)
 * the results are undefined.
 *
 * XXX this call may evolve into a more generic thread_info call.
 */
an_error_t 
an_thread_stackinfo(an_thread_t *thread, void **base, size_t *size, void **ptr)
{
	ani_thread_t *ithread;
	void *ibase, *iptr;
	size_t isize;
	int err;
	
	NODEOS_ENTER;
	ithread = NODEOS_INTERN_THREAD(thread);
	if (ithread == 0)
		err = AN_ERR_INVALIDARG;
	else {
		err = ani_thread_stackinfo(ithread, &ibase, &isize, &iptr);
		if (err == 0)
			NODEOS_COPYOUT(base, &ibase, sizeof ibase, &err);
		if (err == 0)
			NODEOS_COPYOUT(size, &isize, sizeof isize, &err);
		if (err == 0)
			NODEOS_COPYOUT(ptr, &iptr, sizeof iptr, &err);
	}
	NODEOS_LEAVE;

	return NODEOS_EXTERN_ERROR(err);
}
