/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "clock" channel driver.
 *
 * In-channel only.  Packets are delivered at a rate given by the address info.
 * (e.g., if address is "5", a packet is delivered every 5th clock tick).
 * The delivered packet contains the 4-byte internal time value (``mytime'').
 */
#include "options.h"

#ifdef CHAN_CLOCK

#include <sys/time.h>
#ifdef OSKIT
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/dev/clock.h>
#include <oskit/dev/timer.h>
#else
#define oskit_iunknown_t void
#define oskit_error_t int
#endif

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_pbuf.h>

#define MAXCHAN	100
#define FREQ	100
#define CBUFSIZ	sizeof(unsigned int)

#ifdef OSKIT
extern oskit_listener_t *oskit_create_threaded_listener(
			oskit_listener_callback_t handler, void *arg);
extern oskit_clock_t *oskit_system_clock;

static oskit_timer_t *chan_clock_timer;
static oskit_itimerspec_t chan_ts;
#else
static void *chan_clock_clock(void *arg);
#endif

static oskit_error_t chan_clock_intr(oskit_iunknown_t *l, void *arg);

static volatile unsigned int mytime;

static struct chans {
	struct chans *next;
	ani_chan_t *chan;
	int freq;
} chans[MAXCHAN], *allchans, *freechans;
static ani_ilock_t clocklock;

struct clockaddr {
	int freq;
};

void
chan_clock_init(void)
{
#ifdef OSKIT
	oskit_listener_t *l;
#else
	pthread_t id;
	int err;
#endif
	int i;

	ani_ilock_init(&clocklock);
	ani_ilock_lock(&clocklock);

	for (i = 0; i < MAXCHAN; i++) {
		chans[i].next = freechans;
		freechans = &chans[i];
	}

#ifdef OSKIT
	/*
	 * Create a timer with a thread-aware listener.
	 * XXX the handler runs without a flow context.
	 */
	l = oskit_create_threaded_listener(chan_clock_intr, 0);
	oskit_clock_createtimer(oskit_system_clock, &chan_clock_timer);
	oskit_timer_setlistener(chan_clock_timer, l);
	oskit_listener_release(l);
	chan_ts.it_value.tv_sec = 0;
	chan_ts.it_value.tv_nsec = 1000000000 / FREQ;
	chan_ts.it_interval.tv_sec = 0;
	chan_ts.it_interval.tv_nsec = 1000000000 / FREQ;
	oskit_timer_settime(chan_clock_timer, 0, &chan_ts);
#else
	/*
	 * XXX hack: create a thread to just do select-as-timer and invoke
	 * the interrupt handler whenever it wakes up.  We could do this for
	 * the OSKit as well, but select is very heavyweight in the OSKit.
	 */
	err = pthread_create(&id, 0, chan_clock_clock, 0);
	assert(err == 0);
	err = pthread_detach(id);
	assert(err == 0);
#endif

	ani_ilock_unlock(&clocklock);
}

int
chan_clock_create(ani_chan_t *chan)
{
	int freq;
	struct chans *ch;

	freq = chan->addr.clock.freq;
	if (freq < 1)
		return EINVAL;

	ani_ilock_lock(&clocklock);
	ch = freechans;
	if (ch == 0) {
		ani_ilock_unlock(&clocklock);
		return ENOMEM;
	}
	freechans = ch->next;
	ani_ilock_unlock(&clocklock);

	ch->chan = chan;
	ch->freq = freq;
	chan->data = ch;

	return 0;
}

void
chan_clock_destroy(ani_chan_t *chan)
{
	struct chans *ch, *nch;

	assert(chan->data != 0);
	ch = chan->data;

	ani_ilock_lock(&clocklock);
	for (nch = allchans; nch != 0; nch = nch->next)
		assert(ch != nch);
	ch->next = freechans;
	freechans = ch;
	ani_ilock_unlock(&clocklock);

	chan->data = 0;
}

int
chan_clock_enable(ani_chan_t *chan)
{
	struct chans *ch;

	assert(chan->data != 0);
	ch = chan->data;
	assert(ch->freq != 0);

	ani_ilock_lock(&clocklock);
	ch->next = allchans;
	allchans = ch;
	ani_ilock_unlock(&clocklock);

	return 0;
}

void
chan_clock_disable(ani_chan_t *chan)
{
	struct chans *ch, **pch;

	assert(chan->data != 0);

	ch = chan->data;
	ani_ilock_lock(&clocklock);
	for (pch = &allchans; *pch != 0; pch = &(*pch)->next)
		if (*pch == ch) {
			*pch = ch->next;
			break;
		}
	ani_ilock_unlock(&clocklock);
}

static oskit_error_t
chan_clock_intr(oskit_iunknown_t *l, void *arg)
{
	struct chans *ch;
	ani_pbuf_t *pbuf;
	unsigned int ot;
	int err;

	ani_ilock_lock(&clocklock);
	ot = ++mytime;
	for (ch = allchans; ch != 0; ch = ch->next)
		if (ch->freq == 1 || (ot % ch->freq) == 0) {
			ani_chan_lock(ch->chan); /* XXX locking order */
			err = ani_chan_pbuf_alloc(ch->chan, CBUFSIZ, &pbuf);
			if (err == 0) {
				*(unsigned int *)ani_pbuf_data(pbuf) = ot;
				err = ani_chan_thread_start(ch->chan, pbuf);
				if (err != 0)
					ani_chan_pbuf_free(ch->chan, pbuf);
			}
			ani_chan_unlock(ch->chan);
		}
	ani_ilock_unlock(&clocklock);

	return 0;
}

#ifndef OSKIT

#include <unistd.h>

static void *
chan_clock_clock(void *arg)
{
	struct timeval delta;

	delta.tv_sec = 0;

	while (1) {
		delta.tv_usec = 1000000 / FREQ;
		(void) select(0, 0, 0, 0, &delta);
		chan_clock_intr(0, 0);
	}

	return 0;
}

#endif

#endif /* CHAN_CLOCK */
