/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * A "socket" channel driver.
 *
 * Currently we recognize only "socket/udp/ip/eth".
 */
#include "options.h"

#ifdef CHAN_SOCKET

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#ifdef OSKIT
#include <oskit/startup.h>
#include <oskit/com/listener.h>
#include <oskit/dev/dev.h>
#include <oskit/net/socket.h>
#include <oskit/dev/timer.h>
#endif

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_thread.h>
#include "threadpool.h"
#include "demux.h"
#include "mem.h"

/*
 * One structure per socket.
 * Sockets may be shared by multiple channels, with each channel having
 * a different demux for inChans.
 */
struct socketdata {
	struct socketdata  *next;
	short		   stype; /* SOCK_DGRAM or SOCK_STREAM */
	short		   ctype; /* inChan or outChan */
	struct sockaddr	   addr;  /* IP addr and UDP/TCP port */
	int		   fd;	  /* socket file descriptor */
	demux_t		   *demux;/* list of demuxers for this socket */
	int		   refs;  /* refcount */
	ani_ilock_t	   lock;  /* protects the demux list */
};

static struct socketdata *socketlist;
static ani_ilock_t socketlock;

/*
 * Channel specific data (chan->data) is a pointer to a socketdata
 * structure (inChan) or a socket file descriptor (outChan).
 */

static int chan_socketdata_find(int stype, an_sockaddr_t *caddr,
				int ctype, struct socketdata **out_sdata);
static void chan_socketdata_release(struct socketdata *sdata);
static void chan_socketdata_intr(void *arg, void *arg2);

/*
 * In a couple of places we need to create "uncharged" resources.
 * These are accounted against the root flow.
 */
extern ani_flow_t *root_flow;

void
chan_socket_init(void)
{
	ani_ilock_init(&socketlock);

#ifdef OSKIT        
	ani_ilock_lock(&socketlock);
	start_network_router_pthreads();
	ani_ilock_unlock(&socketlock);
#endif
}

int
chan_socket_create(ani_chan_t *chan)
{
	struct socketdata *sdata;
	int err, stype, len;
	char *cp;

	/* XXX */
	if (ani_chan_type(chan) == CHAN_CUT)
		return EINVAL;

	cp = ani_chan_name(chan);
	len = ani_chan_namelen(chan);
	if (strncmp(cp, "socket/udp/ip/eth", len) == 0)
		stype = SOCK_DGRAM;
#if 0
	else if (strncmp(cp, "socket/tcp/ip/eth", len) == 0)
		stype = SOCK_STREAM;
#endif
	else
		return EINVAL;

	/*
	 * Find/create a socket of the appropriate type.
	 * Returns the socket data locked.
	 */
	err = chan_socketdata_find(stype, &chan->addr.socket,
				   ani_chan_type(chan), &sdata);
	if (err)
		return err;

	ani_ilock_unlock(&sdata->lock);

	chan->data = sdata;
	return 0;
}

void
chan_socket_destroy(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);
	sdata = chan->data;

	ani_ilock_lock(&sdata->lock);
	chan_socketdata_release(sdata);
	/* lock got destroyed, no need to unlock. */
}

int
chan_socket_enable(ani_chan_t *chan)
{
	struct socketdata *sdata;
	int err = 0;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		err = ani_chan_demux_add(chan, &sdata->demux);
		ani_ilock_unlock(&sdata->lock);
		if (err && chan->tdata.in.demux != 0) {
#if 0
			ani_demuxkey_destroy(chan->tdata.in.demux);
#endif
			chan->tdata.in.demux = 0;
		}
	}

	return err;
}

void
chan_socket_disable(ani_chan_t *chan)
{
	struct socketdata *sdata;

	assert(chan->data);

	if (ani_chan_type(chan) == CHAN_IN) {
		sdata = chan->data;
		ani_ilock_lock(&sdata->lock);
		ani_chan_demux_remove(chan, &sdata->demux);
		ani_ilock_unlock(&sdata->lock);

		if (chan->tdata.in.demux != 0) {
#if 0
			ani_demuxkey_destroy(chan->tdata.in.demux);
#endif
			chan->tdata.in.demux = 0;
		}
	}
}

void
chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	struct socketdata *sdata;

	sdata = chan->data;
	ani_ilock_lock(&sdata->lock);
	(void) write(sdata->fd, ani_pbuf_data(pbuf), ani_pbuf_size(pbuf));
	ani_ilock_unlock(&sdata->lock);
}

/*
 * Lookup or create a socket corresponding to the desired type/address.
 * Returns 0 on success with the socketdata struct locked.
 */
static int
chan_socketdata_find(int stype, an_sockaddr_t *caddr, int ctype,
		     struct socketdata **out_sdata)
{
	struct socketdata *sdata;
	int sbsize, err;

	/*
	 * Look for an existing socket
	 */
	ani_ilock_lock(&socketlock);
	for (sdata = socketlist; sdata != 0; sdata = sdata->next)
		if (sdata->ctype == ctype && sdata->stype == stype &&
		    memcmp(&sdata->addr, caddr, sizeof sdata->addr) == 0) {
			ani_ilock_lock(&sdata->lock);
			ani_ilock_unlock(&socketlock);
			sdata->refs++;
			*out_sdata = sdata;
			return 0;
		}
	ani_ilock_unlock(&socketlock);

	/*
	 * Otherwise, create a new one
	 */
	sdata = mem_alloc(sizeof *sdata);
	if (sdata == 0)
		return ENOMEM;

	sdata->fd = socket(PF_INET, stype, 0);
	if (sdata->fd < 0) {
		mem_free(sdata, sizeof *sdata);
		return errno;
	}

	/*
	 * Make socket buffer size as large as possible.
	 */
	sbsize = 32 * 1024;
	if (setsockopt(sdata->fd, SOL_SOCKET,
		       ctype == CHAN_IN ? SO_RCVBUF : SO_SNDBUF,
		       &sbsize, sizeof sbsize) < 0) {
		err = errno;
		close(sdata->fd);
		mem_free(sdata, sizeof *sdata);
		return err;
	}

	/*
	 * We initialize and lock the sdata lock so we can use it to
	 * synchronize with any reader thread.  We set the ref count now
	 * so that we can use socketdata_release to cleanup on errors.
	 */
	ani_ilock_init(&sdata->lock);
	ani_ilock_lock(&sdata->lock);

	if (ctype == CHAN_IN) {
		if (bind(sdata->fd, caddr, sizeof *caddr) < 0) {
			err = errno;
			goto error;
		}

		err = threadpool_thread_start(root_flow->tpool, 1,
					      chan_socketdata_intr, sdata, 0,
					      0);
		if (err)
			goto error;
	} else {
		if (connect(sdata->fd, caddr, sizeof *caddr) < 0) {
			err = errno;
			goto error;
		}
	}

	/*
	 * Initialize the rest of the socket data
	 *
	 * Note that we acquire the socketlist lock out of order here.
	 * This is safe since no one else could hold our lock yet and cause
	 * deadlock.
	 */
	sdata->stype = stype;
	sdata->ctype = ctype;
	sdata->addr = *caddr;
	sdata->demux = 0;
	sdata->refs = 1;
	ani_ilock_lock(&socketlock);
	sdata->next = socketlist;
	socketlist = sdata;
	ani_ilock_unlock(&socketlock);

	*out_sdata = sdata;
	return 0;

error:
	close(sdata->fd);
	ani_ilock_unlock(&sdata->lock);
	ani_ilock_deinit(&sdata->lock);
	mem_free(sdata, sizeof *sdata);
	return err;
}

/*
 * Drop a reference to a socketdata struct.
 * Called with the sdata locked.  If no more refs, the sdata is destroyed.
 * Otherwise, the lock is dropped.
 */
static void
chan_socketdata_release(struct socketdata *sdata)
{
	struct socketdata **sdp;

	if (--sdata->refs == 0) {
		/*
		 * If we cannot grab the list lock, do the funky lock dance.
		 * Drop our lock to get the global lock then reaquire ours.
		 * Then check to make sure someone didn't latch on to our
		 * sdata in the meantime.
		 */
		if (ani_ilock_trylock(&socketlock)) {
			ani_ilock_unlock(&sdata->lock);
			ani_ilock_lock(&socketlock);
			ani_ilock_lock(&sdata->lock);
			if (sdata->refs != 0) {
				ani_ilock_unlock(&sdata->lock);
				ani_ilock_unlock(&socketlock);
				return;
			}
		}

		/*
		 * Locate and remove from the list
		 */
		for (sdp = &socketlist; *sdp != 0; sdp = &(*sdp)->next)
			if (*sdp == sdata)
				break;
		assert(*sdp != 0);
		*sdp = sdata->next;
		ani_ilock_unlock(&socketlock);

		/*
		 * Shutdown and destroy the socket.
		 * Note that closing the socket will knock the receive
		 * thread out of its read loop and force it to exit.
		 */
		close(sdata->fd);
		ani_ilock_unlock(&sdata->lock);
		ani_ilock_deinit(&sdata->lock);
		mem_free(sdata, sizeof *sdata);
	} else
		ani_ilock_unlock(&sdata->lock);
}

/*
 * Receiver thread for inChan sockets.
 * Applies the demux filters for the socket to determine a channel to
 * deliver to and fires up a channel thread to deal with it.
 *
 * XXX for TCP support, do listen and accept here.
 */
static void
chan_socketdata_intr(void *arg, void *arg2)
{
	struct socketdata *sdata = arg;
	void *bdata;
	ssize_t bsize, size;
	int s, err;
	ani_pbuf_t *rbuf, *pbuf;
	ani_chan_t *chan;

	/*
	 * Locking forces synchronization with our creator
	 */
	ani_ilock_lock(&sdata->lock);
	s = sdata->fd;
	ani_ilock_unlock(&sdata->lock);

	/*
	 * Create an initial read buffer.  We swap this buffer with
	 * empty ones from the socket's channels when we hand off packets.
	 */
	rbuf = mem_alloc(sizeof *rbuf);
	assert(rbuf != 0);
	err = ani_pbuf_init(rbuf, ani_flow_current(), 0,
			    NODEOS_CHAN_SOCK_BSIZE, 0, NODEOS_CHAN_SOCK_BSIZE);
	assert(err == 0);
	bdata = ani_pbuf_data(rbuf);
	bsize = ani_pbuf_size(rbuf);

	/*
	 * Loop receiving and processing data
	 */
	while (1) {
		chan = 0;

		size = recvfrom(s, bdata, bsize, 0, 0, 0);
		if (size <= 0)
			break;

		/*
		 * Locate the target channel for this data.
		 * If none, just drop it and continue.
		 */
		ani_ilock_lock(&sdata->lock);
		chan = demux_apply(sdata->demux, bdata, size);
		ani_ilock_unlock(&sdata->lock);
		if (chan == 0)
			continue;

		/*
		 * Allocate a packet buffer from the channel.
		 * If none are available, just drop the packet.
		 */
		ani_chan_lock(chan);
		err = ani_chan_pbuf_alloc(chan, NODEOS_CHAN_SOCK_BSIZE, &pbuf);
		if (err != 0) {
			ani_chan_unlock(chan);
			continue;
		}

		/*
		 * Swap the buffer obtained from the channel with our buffer.
		 * The channel gets the full buffer in exchange for one of its
		 * empty ones that we will use for the next read.
		 */
		ani_pbuf_setsize(rbuf, size, 0);
		ani_pbuf_swap(rbuf, pbuf);

		/*
		 * Finally, fire up a handler thread.
		 */
		err = ani_chan_thread_start(chan, pbuf);
		if (err)
			ani_chan_pbuf_free(chan, pbuf);

		ani_chan_unlock(chan);

		/*
		 * Reset the buffer pointer/size since we have a new
		 * physical buffer.
		 */
		bdata = ani_pbuf_data(rbuf);
		bsize = ani_pbuf_size(rbuf);
	}

	if (errno != EINTR && errno != EBADF) {
		struct sockaddr_in *sin;
		sin = (struct sockaddr_in *)&sdata->addr;
		printf("0x%x/%d: receive thread died: %s\n",
		       (unsigned int)ntohl(sin->sin_addr.s_addr),
		       ntohs(sin->sin_port), strerror(errno));
	}

	/*
	 * We have broken out of the recvfrom.  Most likely because the
	 * socket was closed.  Free the read buffer and exit.
	 */
	ani_pbuf_deinit(rbuf);
}

#endif /* CHAN_SOCKET */
