/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <string.h>

#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_demuxkey.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/an_pbuf.h>

#include "demux.h"
#include "threadpool.h"
#include "misc.h"
#include "options.h"

#ifdef CHAN_CLOCK
extern void chan_clock_init(void);
extern int chan_clock_create(ani_chan_t *chan);
extern void chan_clock_destroy(ani_chan_t *chan);
extern int chan_clock_enable(ani_chan_t *chan);
extern void chan_clock_disable(ani_chan_t *chan);
#endif
#ifdef CHAN_ETH
extern void chan_eth_init(void);
extern int chan_eth_create(ani_chan_t *chan);
extern void chan_eth_destroy(ani_chan_t *chan);
extern void chan_eth_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
extern int chan_eth_enable(ani_chan_t *chan);
extern void chan_eth_disable(ani_chan_t *chan);
#endif
#ifdef CHAN_IF
extern void chan_if_init(void);
extern int chan_if_create(ani_chan_t *chan);
extern void chan_if_destroy(ani_chan_t *chan);
extern void chan_if_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
extern int chan_if_enable(ani_chan_t *chan);
extern void chan_if_disable(ani_chan_t *chan);
#endif
#ifdef CHAN_SOCKET
extern void chan_socket_init(void);
extern int chan_socket_create(ani_chan_t *chan);
extern void chan_socket_destroy(ani_chan_t *chan);
extern void chan_socket_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
extern int chan_socket_enable(ani_chan_t *chan);
extern void chan_socket_disable(ani_chan_t *chan);
#endif

/*
 * Protocol-specific processing modules.
 */
static struct chanops allchans[] =
{
#ifdef CHAN_CLOCK
{
	"clock",		chan_clock_init,
	chan_clock_create,	chan_clock_destroy,	0,
	chan_clock_enable,	chan_clock_disable
},
#endif
#ifdef CHAN_ETH
{
	"eth",			chan_eth_init,
	chan_eth_create,	chan_eth_destroy,	chan_eth_send,
	chan_eth_enable,	chan_eth_disable
},
#endif
#ifdef CHAN_IF
{
	"if",			chan_if_init,
	chan_if_create,		chan_if_destroy,	chan_if_send,
	chan_if_enable,		chan_if_disable
},
#endif
#ifdef CHAN_SOCKET
{
	"socket",		chan_socket_init,
	chan_socket_create,	chan_socket_destroy,	chan_socket_send,
	chan_socket_enable,	chan_socket_disable
},
#endif
{	0,	0,	0,	0,	0,	0,	0 }
};

void
ani_chan_setup(void)
{
	struct chanops *ops;

	for (ops = allchans; ops->name; ops++)
		if (ops->init)
			ops->init();

	if_setup();
}

/*
 * ``Creates'' a channel object.
 * It acutally initializes an existing piece of memory (``chan'') to make it
 * a channel object.  The ideal here is that we allocate no memory, the
 * caller provides us with pointers to everything we need.  This includes
 * not only the channel struct, but the proto, addr, and demux structs as
 * well; i.e., anything that is passed by reference.
 *
 * XXX we still allocate channel protocol-specific data at this level.
 */
int
ani_chan_init(ani_chan_t *chan, int ctype, ani_flow_t *flow,
	      an_chanproto_t *proto, an_chanaddr_t *addr,
	      ani_demuxkey_t *demux, an_cpuspec_t *cpu,
	      ani_pbuf_t **pbufs, int npbufs, an_netspec_t *net,
	      an_chanrecvfunc_t func, void *farg)
{
	struct chanops *ops;
	int err;

	assert(ctype != CHAN_CUT);

	/*
	 * See if we know how to handle the desired protocol processing.
	 * XXX assumes protocol data is a string.
	 */
	for (ops = allchans; ops->name; ops++)
		if (strncmp(proto->data, ops->name, strlen(ops->name)) == 0)
			break;
	if (ops->name == 0)
		return EINVAL;

	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX CPU/memory resources inherited from the flow right now
	 */
	chan->tpool = chan->flow->tpool;
	chan->mpool = flow->mpool;


	/*
	 * Record the resource specifications.
	 * It is up to the entry layer to perform any cross-address space
	 * copyins, we assume the specs passed to us are directly accessible.
	 */

	/*
	 * XXX right now we just store the raw, user supplied form of the
	 * address and protocol specs.  In the future we might want to reduce
	 * and compact this info sometime.
	 */
	if (addr != 0)
		memcpy(&chan->addr, addr, sizeof chan->addr);
	if (proto != 0)
		memcpy(&chan->proto, proto, sizeof chan->proto);
	/*
	 * XXX for CPU we only care about max threads right now.
	 */
	if (cpu != 0)
		chan->maxthreads = cpu->max;

	chan->ops = ops;
	chan->flags = ctype | CHAN_ISRAW;
	chan->refs = 1;
	ani_ilock_init(&chan->lock);

	/*
	 * Do channel-type specific setup
	 */
	if (ctype == CHAN_IN) {
		/*
		 * XXX make sure we always have a demux object for the
		 * benefit of channel types that need to augment the demux.
		 *
		 * XXX this memory never gets freed right now.
		 */
		if (demux == 0) {
			demux = ani_mempool_alloc(chan->mpool, sizeof *demux);
			if (demux == 0) {
				ani_ilock_deinit(&chan->lock);
				return ENOMEM;
			}
			err = ani_demuxkey_init(demux);
			assert(err == 0);
		}
		chan->tdata.in.demux = demux;
		chan->tdata.in.func = func;
		chan->tdata.in.farg = farg;

		if (npbufs) {
			err = ani_chan_pbuf_add(chan, pbufs, npbufs);
			if (err != 0) {
				ani_ilock_deinit(&chan->lock);
				return err;
			}
		}
	}

	/*
	 * Do channel-protocol specific setup
	 */
	ani_chan_lock(chan);
	err = ops->create(chan);
	if (err) {
		ani_chan_unlock(chan);
		ani_ilock_deinit(&chan->lock);
		return err;
	}

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 */
	ani_chan_unlock(chan);
	ani_flow_addchannel(flow, chan);

	return 0;
}

static void
ani_chan_deinit_internal(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));

	/*
	 * Note that the channel is dying.  Needed in case threads block
	 * on the channel lock while we are destroying the channel.
	 */
	chan->flags |= CHAN_DYING;

	/*
	 * Make sure the channel is disabled to avoid races with receiver
	 * threads.
	 */
	if (ani_chan_type(chan) == CHAN_IN) {
		chan->ops->disable(chan);

		/*
		 * Once the channel is disabled and marked as dying,
		 * we can free up in-channel packet buffers.
		 */
		ani_chan_pbuf_removeall(chan);
	}

	/*
	 * For cut "pseudo" channels, we destroy the constituent channels.
	 * Otherwise we call the channel-protocol specific shutdown.
	 */
	if (ani_chan_type(chan) == CHAN_CUT) {
		ani_chan_t *cchan;

		cchan = chan->tdata.cut.inchan;
		assert(chan != cchan);
		ani_chan_lock(cchan);
		ani_chan_deinit_internal(cchan);
		ani_chan_unlock(cchan);
		ani_chan_release(cchan);

		cchan = chan->tdata.cut.outchan;
		assert(chan != cchan);
		ani_chan_lock(cchan);
		ani_chan_deinit_internal(cchan);
		ani_chan_unlock(cchan);
		ani_chan_release(cchan);
	} else
		chan->ops->destroy(chan);
}

int
ani_chan_deinit(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	/*
	 * XXX hack, don't allow explicit destruction of cut-channel
	 * components.
	 */
	if (chan->flags & CHAN_WITHCUT)
		return EINVAL;

	/*
	 * Remove from the flow before we lock the channel
	 */
	ani_flow_removechannel(chan->flow, chan);

	ani_chan_lock(chan);
	ani_chan_deinit_internal(chan);
	ani_chan_unlock(chan);
	ani_chan_release(chan);

	return 0;
}

void
ani_chan_cleanup(ani_chan_t *chan)
{
	assert((chan->flags & CHAN_DYING) != 0);
	assert(ani_chan_islocked(chan) == 0);

#if 0
	CHANSTATS(ani_chan_stats_dump(chan));
#endif
	ani_ilock_deinit(&chan->lock);
}

int
ani_chan_enable(ani_chan_t *chan)
{
	int err;

	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);
	if (ani_chan_type(chan) == CHAN_IN && chan->tdata.in.func != 0)
		err = chan->ops->enable(chan);
	else
		err = EINVAL;
	ani_chan_unlock(chan);

	return err;
}

void
ani_chan_disable(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);
	chan->ops->disable(chan);
	ani_chan_unlock(chan);
}

int
ani_chan_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	assert_ani_chan_valid(chan);

	if (ani_chan_type(chan) != CHAN_OUT)
		return EINVAL;
	if (chan->flags & CHAN_DYING)
		return EBADF; /* XXX */

	ani_chan_lock(chan);
	if (chan->ops->send) {
		chan->ops->send(chan, pbuf);
		CHANSTATS(chan->stats.outpackets++);
		CHANSTATS(chan->stats.outbytes += ani_pbuf_size(pbuf));
	}
	ani_chan_unlock(chan);

	return 0;
}

an_chanstats_t *
ani_chan_stats(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	if (ani_chan_type(chan) == CHAN_CUT) {
		/* XXX copy stats out of constituent channels */
		chan->stats.inpackets = chan->tdata.cut.inchan->stats.inpackets;
		chan->stats.inbytes = chan->tdata.cut.inchan->stats.inbytes;
		chan->stats.droppackets =
			chan->tdata.cut.inchan->stats.droppackets;
		chan->stats.dropnobufs =
			chan->tdata.cut.inchan->stats.dropnobufs;
		chan->stats.dropfailedthread =
			chan->tdata.cut.inchan->stats.dropfailedthread;
		chan->stats.dropfailedbuf =
			chan->tdata.cut.inchan->stats.dropfailedbuf;
		chan->stats.bufcopies = chan->tdata.cut.inchan->stats.bufcopies;
		chan->stats.bufqueued = chan->tdata.cut.inchan->stats.bufqueued;
		chan->stats.maxthreads =
			chan->tdata.cut.inchan->stats.maxthreads;
		chan->stats.outpackets =
			chan->tdata.cut.outchan->stats.outpackets;
		chan->stats.outbytes = chan->tdata.cut.outchan->stats.outbytes;
	}
	return &chan->stats;
}

int
ani_chan_thread_start(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	int err, inkern;
	void *pktarg;

	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));
	assert(chan->tdata.in.func != 0);

	/*
	 * If channel is at its thread limit, queue the packet for later
	 */
	if (chan->curthreads >= ani_chan_maxthreads(chan)) {
		ani_pbuf_append(&chan->tdata.in.qbufs, pbuf);
		CHANSTATS(chan->stats.bufqueued++);
		return 0;
	}

	/*
	 * XXX special case for in side of a cut-through channel.
	 *
	 * The handler function for a cut-channel is run in the NodeOS
	 * and is passed an internal pbuffer object.
	 *
	 * The handler function for an in-channel is run in the EE and
	 * is passed the external handle for a packet.
	 */
	if (chan->flags & CHAN_WITHCUT) {
		inkern = 1;
		pktarg = pbuf;
	} else {
		inkern = 0;
		pktarg = NODEOS_EXTERN_PBUF(pbuf);
	}

	err = threadpool_thread_schedule(chan->tpool, chan, inkern, 0,
					 (thread_func_t)chan->tdata.in.func,
					 pktarg, chan->tdata.in.farg);
	if (err == 0) {
		CHANSTATS(chan->stats.inpackets++);
		CHANSTATS(chan->stats.inbytes += ani_pbuf_size(pbuf));
		ani_chan_reference(chan);
		chan->curthreads++;
#ifdef CHANNEL_STATS
		if (chan->curthreads > chan->stats.maxthreads)
			chan->stats.maxthreads = chan->curthreads;
#endif
	} else {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedthread++);
	}

	return err;
}

/*
 * Called from thread_loop when a thread is done processing a channel
 * packet.  If there were pending packets for this channel, put one of
 * them on the worklist for this thread to pick up on return.
 */
void
ani_chan_thread_exit(ani_chan_t *chan)
{
	assert_ani_chan_valid(chan);

	ani_chan_lock(chan);
	if (chan->tdata.in.qbufs != 0) {
		int err, inkern;
		void *pktarg;
		ani_pbuf_t *pbuf;

		pbuf = ani_pbuf_dequeue(&chan->tdata.in.qbufs);

		/*
		 * XXX special case for in side of a cut-through channel.
		 *
		 * The handler function for a cut-channel is run in the NodeOS
		 * and is passed an internal pbuffer object.
		 *
		 * The handler function for an in-channel is run in the EE and
		 * is passed the external handle for a packet.
		 */
		if (chan->flags & CHAN_WITHCUT) {
			inkern = 1;
			pktarg = pbuf;
		} else {
			inkern = 0;
			pktarg = NODEOS_EXTERN_PBUF(pbuf);
		}
		err = threadpool_thread_schedule(chan->tpool, chan, inkern, 1,
					(thread_func_t)chan->tdata.in.func,
					pktarg, chan->tdata.in.farg);
		assert(err == 0);
		CHANSTATS(chan->stats.inpackets++);
		CHANSTATS(chan->stats.inbytes += ani_pbuf_size(pbuf));
		ani_chan_unlock(chan);
	} else {
		--chan->curthreads;
		ani_chan_unlock(chan);
		ani_chan_release(chan);
	}
}

/*
 * Channel packet buffer handling.
 */

int
ani_chan_pbuf_alloc(ani_chan_t *chan, size_t size, ani_pbuf_t **out_pbuf)
{
	ani_pbuf_t *pbuf;
	int err;

	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(size >= 0);
	assert(out_pbuf != 0);

	/*
	 * Grab a buffer of the appropriate size.
	 */
	pbuf = ani_pbuf_dequeue(&chan->tdata.in.rbufs);
	if (pbuf == 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropnobufs++);
		return ENOMEM;
	}
	assert(ani_pbuf_bufsize(pbuf) >= size);
	err = ani_pbuf_setsize(pbuf, size, 0);
	assert(err == 0);

	*out_pbuf = pbuf;
	return 0;
}

/*
 * Place a buffer back on the channel's buffer list.
 * Channel is locked on entry.
 */
void
ani_chan_pbuf_free(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	assert_ani_chan_valid(chan);
	assert(ani_chan_islocked(chan));
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(pbuf != 0);

	/*
	 * If the channel is in the process of dying just free the buffer
	 */
	if (chan->flags & CHAN_DYING)
		ani_pbuf_deinit(pbuf);
	else {
		/*
		 * XXX we ignore any EE specified offset/size initializations
		 * as there is very little we can do with them without copying
		 * every packet.
		 */
		ani_pbuf_setmax(pbuf);

		ani_pbuf_enqueue(&chan->tdata.in.rbufs, pbuf);
	}
}

/*
 * Associate a list of pbufs with a channel.
 * Channel is unlocked on call and return.
 *
 * This is so that this routine can be used to add buffering to an existing
 * channel.  In that case, we don't want to keep an active channel locked
 * while we allocate memory.
 */
int
ani_chan_pbuf_add(ani_chan_t *chan, ani_pbuf_t *pbufs[], int npbufs)
{
	ani_pbuf_t *fpb;
	int i;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(pbufs != 0 || npbufs == 0);

	/*
	 * Chain the new bufs together.
	 */
	fpb = 0;
	for (i = 0; i < npbufs; i++) {
		assert(pbufs[i] != 0);

		/*
		 * XXX we ignore any EE specified offset/size initializations
		 * as there is very little we can do with them without copying
		 * every packet.
		 */
		ani_pbuf_setmax(pbufs[i]);

		ani_pbuf_enqueue(&fpb, pbufs[i]);

		/* XXX */
		pbufs[i]->chan = chan;
	}

	/*
	 * Now add them to the channel's list.
	 * This is done is a separate step to avoid holding the channel
	 * lock too long.
	 *
	 * XXX need to make sure channel hasn't died while we were
	 * allocating memory above.
	 */
	ani_chan_lock(chan);
	ani_pbuf_append(&chan->tdata.in.rbufs, fpb);
	ani_chan_unlock(chan);

	return 0;
}

/*
 * Remove and free all buffers associated with a channel.
 * Called during channel destroy.  Channel should be locked on call and return.
 *
 * @deallocation: all buffers on the channel list
 */
void
ani_chan_pbuf_removeall(ani_chan_t *chan)
{
	ani_pbuf_t *pb;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);
	assert(ani_chan_islocked(chan));

	/*
	 * Loop through, freeing all the buffers.
	 * Time spent holding the lock is not an issue here
	 * since the channel is dying anyway.
	 */
	while ((pb = ani_pbuf_dequeue(&chan->tdata.in.qbufs)) != 0)
		ani_pbuf_deinit(pb);
	while ((pb = ani_pbuf_dequeue(&chan->tdata.in.rbufs)) != 0)
		ani_pbuf_deinit(pb);
}


/*
 * Channel demux handling
 */

/*
 * Create and install a demux object on the given demux list.
 * Channel should be locked on call.
 */
int
ani_chan_demux_add(ani_chan_t *chan, struct demux **listp)
{
	demux_t *demux;
	int prio, err;

	assert_ani_chan_valid(chan);
	assert(ani_chan_type(chan) == CHAN_IN);

	if (chan->tdata.in.demux) {
		err = demux_create(chan, ani_demuxkey_match,
				   chan->tdata.in.demux, &demux);
		prio = chan->tdata.in.demux->ntriples; /* XXX */
	} else {
		err = demux_create(chan, 0, 0, &demux);
		prio = 0; /* XXX */
	}

	if (err == 0) {
		err = demux_add(listp, demux, prio);
		if (err != 0)
			demux_destroy(demux);
	}

	return err;
}

/*
 * Remove a demux object from the specified list.
 * Channel should be locked on call
 */
void
ani_chan_demux_remove(ani_chan_t *chan, struct demux **listp)
{
	demux_t *demux;

	assert_ani_chan_valid(chan);

	demux = demux_find(*listp, chan);
	if (demux) {
		demux_remove(listp, demux);
		demux_destroy(demux);
	}
}


/*
 * Cut-through channel support
 */

/*
 * Create a cutchan given an existing in and out chan.
 */
int
ani_cutchan_init(ani_chan_t *chan, ani_flow_t *flow, an_chanproto_t *proto,
		 ani_chan_t *inchan, ani_chan_t *outchan,
		 an_chanrecvfunc_t rfunc)
{
	int err;

	memset(chan, 0, sizeof *chan);

	/*
	 * XXX set this now so that functions we call know who to charge
	 * for memory.  We officially link into the flow after everything
	 * has been setup.
	 */
	chan->flow = flow;

	/*
	 * XXX CPU/memory resources inherited from the flow right now
	 */
	chan->tpool = flow->tpool;
	chan->mpool = flow->mpool;

	if (proto != 0) {
		NODEOS_COPYIN(&chan->proto, proto, sizeof chan->proto, &err);
		if (err != 0)
			return err;
	}
	chan->flags = CHAN_CUT;
	chan->tdata.cut.inchan = inchan;
	chan->tdata.cut.outchan = outchan;
	chan->refs = 1;
	ani_ilock_init(&chan->lock);

	ani_chan_lock(chan);

	/* XXX setup protocol processing */
	assert(proto == 0);

	/*
	 * Setup the inchan's recv function.
	 * Function argument is the cut channel object.
	 */
	assert(ani_chan_type(inchan) == CHAN_IN);
	inchan->flags |= CHAN_WITHCUT;
	inchan->tdata.in.func = rfunc;
	inchan->tdata.in.farg = chan;

	/*
	 * Mark the outchan as part of the cutchan.
	 */
	assert(ani_chan_type(outchan) == CHAN_OUT);
	outchan->flags |= CHAN_WITHCUT;

	/*
	 * Finally link to the flow.
	 * We drop the channel lock first to adhere to our locking order.
	 * Note that we remove the component in/out channels from the flow's
	 * list so they are not accessible by the EE any longer.  This is
	 * more of a hack than anything, but it simplifies the semantics
	 * a little--don't have to worry about what happens if someone
	 * tweaks the in/out channel while the cutchan is active.
	 *
	 * XXX argh!  Have to leave the channels pointing to the flow so
	 * that foo_current() calls work; i.e., for resource accounting. 
	 */
	ani_chan_unlock(chan);
	ani_flow_removechannel(flow, inchan);
	inchan->flow = flow;
	ani_flow_removechannel(flow, outchan);
	outchan->flow = flow;
	ani_flow_addchannel(flow, chan);

	return 0;
}

/*
 * Process an incoming packet on the inchan-side of a cutchan.
 */
void
ani_cutchan_receive(an_pbuf_t *anpbuf, void *arg)
{
	ani_chan_t *chan, *inchan, *outchan;
	ani_pbuf_t *pbuf;

	chan = arg;
	assert_ani_chan_valid(chan);
	ani_chan_lock(chan);
	if (chan->flags & CHAN_DYING) {
		ani_chan_unlock(chan);
		return;
	}

	assert(ani_chan_type(chan) == CHAN_CUT);
	inchan = chan->tdata.cut.inchan;
	assert_ani_chan_valid(inchan);
	outchan = chan->tdata.cut.outchan;
	assert_ani_chan_valid(outchan);

	pbuf = (ani_pbuf_t *)anpbuf;
	assert(pbuf != 0);

	/* XXX do cutchan protocol processing */

#if 0
	{
		unsigned int *ip = ani_pbuf_data(pbuf);

		printf("%p: %d: %#x/%#x/%#x/%#x\n", chan, ani_pbuf_size(pbuf),
		       ntohl(ip[0]), ntohl(ip[1]), ntohl(ip[2]), ntohl(ip[3]));
	}
#endif

	(void)ani_chan_send(outchan, pbuf);

	ani_chan_lock(inchan);
	ani_chan_pbuf_free(inchan, pbuf);
	ani_chan_unlock(inchan);

	ani_chan_unlock(chan);
}

#ifdef CHANNEL_STATS
#include <stdio.h>

void
ani_chan_stats_dump(ani_chan_t *chan)
{
	/* XXX reported as part of cutchan */
	if (chan->flags & CHAN_WITHCUT)
		return;

	/* XXX pull up stats for cut-channel */
	if (ani_chan_type(chan) == CHAN_CUT)
		(void) ani_chan_stats(chan);

	if (chan->stats.inpackets == 0 && chan->stats.outpackets == 0 &&
	    chan->stats.droppackets == 0)
		return;

	printf("CH%p packets: %d in, %d out, %d queued, %d dropped, %d copied\n",
	       chan, chan->stats.inpackets, chan->stats.outpackets,
	       chan->stats.bufqueued, chan->stats.droppackets,
	       chan->stats.bufcopies);
	printf("\tdrops: %d no bufs, %d/%d failed thread/buf\n",
	       chan->stats.dropnobufs, chan->stats.dropfailedthread,
	       chan->stats.dropfailedbuf);
	printf("\tbytes: %u in, %u out\n",
	       chan->stats.inbytes, chan->stats.outbytes);
	printf("\tmax threads: %d\n", chan->stats.maxthreads);
}
#endif
