/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */
#include <string.h>
#include <assert.h>
#include <stdio.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_demuxkey.h>
#include "demux.h"
#include "mem.h"

/*
 * Define to print info about demux matching
 */
#undef DEMUX_MATCH_DUMP

struct demux_triple {
	int off;
	int len;
	unsigned char key[0];
};

static struct demux_triple *demux_triple_create(int off, int len, void *key);
static void demux_triple_destroy(struct demux_triple *dt);
static void demux_triple_dump(struct demux_triple *dt);

int
ani_demuxkey_init(ani_demuxkey_t *list)
{
	assert(list != 0);

	memset(list, 0, sizeof *list);
	list->listsize = ANI_DEMUXKEY_MAX;

	return 0;
}

int
ani_demuxkey_deinit(ani_demuxkey_t *list)
{
	int i;

	assert(list != 0);

	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		demux_triple_destroy(list->list[i]);
	}
	list->ntriples = 0;

	return 0;
}

int
ani_demuxkey_clone(ani_demuxkey_t *newlist, ani_demuxkey_t *list)
{
	int i;

	assert(newlist != 0);
	assert(list != 0);

	newlist->maxoff = list->maxoff;
	newlist->listsize = list->listsize;
	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		newlist->list[i] = demux_triple_create(list->list[i]->off,
						       list->list[i]->len,
						       list->list[i]->key);
		if (newlist->list[i] == 0) {
			ani_demuxkey_deinit(newlist);
			return ENOMEM;
		}
		newlist->ntriples++;
	}

	return 0;
}

/*
 * Create and add a match element to the front of the given list.
 */
int
ani_demuxkey_prepend(ani_demuxkey_t *list, int off, int len, void *key)
{
	struct demux_triple *dt;
	int i;

	assert(list != 0);

	i = list->ntriples;
	if (i == list->listsize)
		return ENOMEM;

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ENOMEM;

	while (i > 0) {
		list->list[i] = list->list[i-1];
		i--;
	}
	list->list[i] = dt;
	list->ntriples++;

	if (off + len > list->maxoff)
		list->maxoff = off + len;

	return 0;
}


/*
 * Create and add a match element to the end of the given list.
 */
int
ani_demuxkey_append(ani_demuxkey_t *list, int off, int len, void *key)
{
	struct demux_triple *dt;
	int i;

	assert(list != 0);

	i = list->ntriples;
	if (i == list->listsize)
		return ENOMEM;

	dt = demux_triple_create(off, len, key);
	if (dt == 0)
		return ENOMEM;

	list->list[i] = dt;
	list->ntriples++;

	if (off + len > list->maxoff)
		list->maxoff = off + len;

	return 0;
}

/*
 * Adjust the offset of all triples by the indicated amount
 */
void
ani_demuxkey_adjust(ani_demuxkey_t *list, int adjust)
{
	int i;

	assert(list != 0);

	if (list->ntriples != 0) {
		for (i = 0; i < list->ntriples; i++) {
			assert(list->list[i] != 0);
			list->list[i]->off += adjust;
		}
		list->maxoff += adjust;
	}
}

/*
 * Determine if the list has room for the given number of new elements.
 * Returns 0 if so, an error otherwise.
 */
int
ani_demuxkey_expand(ani_demuxkey_t *list, int nelts)
{
	assert(list != 0);

	if (list->ntriples + nelts > list->listsize)
		return ENOMEM;

	return 0;
}

/*
 * Demux match function for a triple list
 */
int
ani_demuxkey_match(void *ddata, void *pdata, int plen)
{
	ani_demuxkey_t *list = ddata;
	struct demux_triple *dt;
	int i;

	assert(list != 0);

	if (list->maxoff > plen)
		return 0;

#ifdef DEMUX_MATCH_DUMP
	{
		int j;
		unsigned char *cp = pdata;

		printf("DATA: ");
		for (j = 0; j < list->maxoff; j++)
			printf("%02x", cp[j]);
		printf("\n");
	}
#endif
	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		dt = list->list[i];
#ifdef DEMUX_MATCH_DUMP
		{
			int j, len;
			unsigned char *cp = pdata + dt->off;

			demux_triple_dump(dt);
			printf("DATA: ");
			len = plen - dt->off;
			if (len > dt->len)
				len = dt->len;
			for (j = 0; j < len; j++)
				printf("%02x", cp[j]);
			printf("\n");
		}
#endif

		if (memcmp(pdata+dt->off, dt->key, dt->len) != 0) {
#ifdef DEMUX_MATCH_DUMP
			printf("FAILED\n");
#endif
			return 0;
		}
	}

#ifdef DEMUX_MATCH_DUMP
	printf("MATCHED\n");
#endif
	return 1;
}

static struct demux_triple *
demux_triple_create(int off, int len, void *key)
{
	struct demux_triple *dt;

	dt = mem_alloc(sizeof *dt + len);
	if (dt) {
		dt->off = off;
		dt->len = len;
		if (key)
			memcpy(dt->key, key, len);
		else
			memset(dt->key, 0, len);
	}

	return dt;
}

static void
demux_triple_destroy(struct demux_triple *dt)
{
	mem_free(dt, sizeof *dt + dt->len);
}

static void
demux_triple_dump(struct demux_triple *dt)
{
	int j;

	printf("TRIPLE %p: %d/%d/0x", dt, dt->off, dt->len);
	for (j = 0; j < dt->len; j++)
		printf("%02x", dt->key[j]);
	printf("\n");
}

void
ani_demuxkey_dump(ani_demuxkey_t *list)
{
	int i;

	printf("TLIST %p, ntriples=%d, maxoff=%d:\n",
	       list, list->ntriples, list->maxoff);
	for (i = 0; i < list->ntriples; i++) {
		assert(list->list[i] != 0);
		demux_triple_dump(list->list[i]);
	}
}
