/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdlib.h>
#include <assert.h>

#include <nodeos/an_resource.h>
#include <nodeos/an_mempool.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_thread.h>
#include <nodeos/impl/ani_nodeos.h>
#include <nodeos/impl/ani_cred.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include "threadpool.h"
#include "mem.h"

static ani_flow_t root_flow_data;
static int ani_flow_destroy_internal(struct hnode *h, void *arg);
static void ani_flow_upcall(ani_flow_t *fflow, ani_flow_t *tflow);

ani_flow_t *root_flow = &root_flow_data;

extern int bootstrapped;

void
ani_flow_setup(an_flowinit_func_t initfunc, an_flowterm_func_t termfunc)
{
	extern ani_cred_t *root_cred;
	extern ani_mempool_t *root_mempool;
	an_cpuspec_t cspec;
	an_memspec_t mspec;
	int err;

	cspec.max = 8; /* XXX */
	cspec.rate = RESOURCE_ALL;
	cspec.slice.tv_sec = RESOURCE_ALL;
	cspec.sched = 0; /* XXX */
	cspec.stacksize = 64 * 1024;
	mspec.npages = 100; /* XXX */
	mspec.pool = NODEOS_EXTERN_MEMPOOL(root_mempool);
	err = ani_flow_init(root_flow, root_cred, &cspec, &mspec, initfunc,
			    0, termfunc);
	assert(err == 0);
}

ani_flow_t *
ani_flow_current(void)
{
	ani_thread_t *thread = ani_thread_current();

	return thread ? thread->pool->flow : 0;
}

int
ani_flow_init(ani_flow_t *flow, ani_cred_t *cred,
	      an_cpuspec_t *cspec, an_memspec_t *mspec,
	      an_flowinit_func_t initfunc, void *initarg,
	      an_flowterm_func_t termfunc)
{
	ani_flow_t *pflow;
	threadpool_t *tpool;
	ani_mempool_t *mpool;
	resource_t res;
	int err;

	/* Sanity */
	if (flow == 0 || cred == 0 || cspec == 0 || mspec == 0 ||
	    initfunc == 0 || termfunc == 0)
		return EINVAL;

	/* Can this flow create sub flows? */
	if (ani_cred_check(cred, CCHECK_FLOW_CREATE) == CCHECK_FAIL)
		return EPERM;

	/* Ensure requested resources are available and reserve them */
	resource_init(&res, cspec, mspec);
	err = ani_cred_resource_alloc(cred, &res);
	if (err)
		return err;

	pflow = ani_flow_current();

	/* Extract the target mempool so we can start charging it */
	mpool = NODEOS_INTERN_MEMPOOL(mspec->pool);
	if (mpool == 0) {
		ani_cred_resource_free(cred, &res);
		return EINVAL;
	}

	/* Initialize the flow */
	hnode_init(&flow->me);
	flow->cred = cred;
	flow->res = res;
	flow->chanlist = 0;
	flow->flags = 0;
	flow->termfunc = termfunc;
	flow->refs = 1;
	ani_ilock_init(&flow->lock);

	/* Link to the mempool now for threadpool memory allocations */
	flow->mpool = mpool;
	ani_mempool_addflow(flow->mpool, flow);

	/* Allocate and initialize the flow threadpool */
	err = threadpool_create(flow, 0, cspec, &tpool);
	if (err) {
		ani_mempool_removeflow(mpool, flow);
		ani_ilock_deinit(&flow->lock);
		ani_cred_resource_free(cred, &res);
		return err;
	}
	flow->tpool = tpool;

	/* Link into flow hierarchies (special handling for first flow) */
	if (flow == root_flow) {
		ani_flow_lock(flow);
		hnode_attach(&flow->me, 0, 0);
	} else {
		ani_flow_lock(root_flow);
		ani_flow_lock(flow);
		hnode_attach(&flow->me, &ani_flow_current()->me, 0);
		ani_flow_unlock(root_flow);
	}

	ani_flow_unlock(flow);

	/* All setup, fire up a thread in the flow to call the init function */
	if (initfunc) {
		err = ani_flow_startthread(flow, 0,
					   (void (*)(void *, void *))initfunc,
					   initarg, 0);
		if (err) {
			ani_flow_deinit(flow);
			return err;
		}
	}

	return 0;
}

int
ani_flow_deinit(ani_flow_t *flow)
{
	ani_flow_t *oflow;
	int me = 0, err;

	assert_ani_flow_valid(flow);
	assert(flow != root_flow);

	/*
	 * Flow must be us or one of our immediate children
	 */
	if (flow == ani_flow_current())
		me = 1;
	else {
		for (oflow = ani_flow_child(ani_flow_current()); oflow != 0;
		     oflow = ani_flow_sibling(oflow))
			if (oflow == flow)
				break;
		if (oflow == 0)
			return EINVAL;
	}

	/*
	 * Prevent recursive upcalls and, effectively, attempts to destroy
	 * an already dying flow.  Note that the flow lock is not held while
	 * in the upcall so we need this other mechanism to ensure this.
	 *
	 * XXX we need to do this better however since there is a race from
	 * here to when the upcall is actually made.
	 */
	if (flow->flags & FLOW_INUPCALL)
		return 0;

	oflow = ani_flow_parent(flow);
	ani_flow_upcall(flow, oflow);

	if (flow != root_flow)
		ani_flow_lock(root_flow);

	err = hnode_traverse(&flow->me, ani_flow_destroy_internal, 0,
			     HNODE_PARENT, 1);
	assert(err == 0);

	if (flow != root_flow)
		ani_flow_unlock(root_flow);

	/* If we just destroyed ourself, no where to return to */
	if (me) {
		/* XXX what resources are we running on? */
		ani_thread_exit();
	}

	return 0;
}

/*
 * The guts of flow destruction.  Called by the hnode_traverse in
 * ani_flow_destroy.
 */
static int
ani_flow_destroy_internal(struct hnode *h, void *arg)
{
	ani_flow_t *flow = (ani_flow_t *)h;
	ani_flow_t *pflow;
	ani_chan_t *chan, *next;

	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);
	pflow = ani_flow_parent(flow);
	flow->flags |= FLOW_DYING;

	/* Remove from flow hierarchy */
	hnode_detach(h, HNODE_PARENT);

	/* Destroy all the channels */
	for (chan = flow->chanlist; chan != 0; chan = next) {
		next = chan->next;
		ani_chan_deinit(chan);
	}

	/* Kill any remaining threads */
	if (flow->tpool)
		threadpool_destroy(flow->tpool);

	/* Release our reference to the flow */
	ani_flow_unlock(flow);
	ani_flow_release(flow);

	return 0;
}

/*
 * Called when there are no more references to the flow.
 * At this point we can safely free resources.
 */
void
ani_flow_cleanup(ani_flow_t *flow)
{
	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);

	/* Remove from mempool */
	assert(hnode_listhead(&flow->me) != 0);
	ani_mempool_removeflow(flow->mpool, flow);

	/* Return resources to the system */
	ani_cred_resource_free(flow->cred, &flow->res);

	/* Uninitialize and free all the data structures */
	ani_flow_unlock(flow);
	ani_ilock_deinit(&flow->lock);
}

/*
 * Upcall to a flow's parent termination handler.
 * If there are no resources to make the call (or the call takes too long)
 * that is fine, we will just forcibly kill everything. 
 */
static void
ani_flow_upcall(ani_flow_t *flow, ani_flow_t *parent)
{
	assert_ani_flow_valid(flow);

	if (flow->termfunc == 0)
		return;

	flow->flags |= FLOW_INUPCALL; /* XXX races */

	assert(parent->tpool != 0);
	(void) threadpool_thread_run(parent->tpool, 0,
				     (thread_func_t)flow->termfunc, flow, 0, 0);
}

void
ani_flow_addchannel(ani_flow_t *flow, ani_chan_t *chan)
{
	assert_ani_flow_valid(flow);

	ani_flow_lock(flow);
	ani_chan_lock(chan);

	/* XXX make sure requested channel resources are consistent with flow */

	chan->flow = flow;
	chan->next = flow->chanlist;
	flow->chanlist = chan;

	ani_chan_unlock(chan);
	ani_flow_unlock(flow);
}

void
ani_flow_removechannel(ani_flow_t *flow, ani_chan_t *chan)
{
	ani_chan_t *ch, **pch;

	assert_ani_flow_valid(flow);
	assert(chan->flow == flow);

	/*
	 * Flows never return from the dead so we don't lock this access
	 */
	if (flow->flags & FLOW_DYING)
		return;

	ani_flow_lock(flow);
	ani_chan_lock(chan);

	for (pch = &flow->chanlist; (ch = *pch) != 0; pch = &ch->next)
		if (ch == chan) {
			*pch = ch->next;
			break;
		}
	chan->flow = 0;

	ani_chan_unlock(chan);
	ani_flow_unlock(flow);
}

int
ani_flow_startthread(ani_flow_t *flow, int inkern,
		     void (*func)(void *, void *), void *farg1, void *farg2)
{
	return threadpool_thread_start(flow->tpool, inkern,
				       (thread_func_t)func, farg1, farg2, 0);
}

#include <stdio.h>

static int
dump_flow(struct hnode *node, void *arg)
{
	ani_flow_t *flow = (ani_flow_t *)node;

	printf("FLOW %p: parent=%p, child=%p, sibling=%p, "
	       "cred=%p, mpool=%p, tpool=%p\n",
	       flow, hnode_parent(node), hnode_child(node), hnode_sibling(node),
	       flow->cred, flow->mpool, flow->tpool);
	return 0;
}

void
ani_flow_dump(ani_flow_t *flow)
{
	if (flow == 0)
		flow = ani_flow_current();

	hnode_traverse(&flow->me, dump_flow, 0, HNODE_PARENT, 0);
}
