/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

/*
 * Network interface information.
 */
#include <assert.h>
#include <stdlib.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_chan.h>
#include "os.h"
#include "options.h"
#include "misc.h"

#ifdef CHAN_ETH
extern void chan_eth_init(void);
extern int chan_eth_create(ani_chan_t *chan);
extern void chan_eth_destroy(ani_chan_t *chan);
extern void chan_eth_send(ani_chan_t *chan, ani_pbuf_t *pbuf);
extern int chan_eth_enable(ani_chan_t *chan);
extern void chan_eth_disable(ani_chan_t *chan);
#endif

static an_ifinfo_t ifinfo[NODEOS_MAXIFS];
static int numifs;

void
if_setup(void)
{
	int err, i;
	an_ifinfo_t *infop;
	static int done;

	if (done++)
		return;

	infop = ifinfo;
	err = 0;

#ifdef CHAN_ETH
	if (err == 0) {
		int count;
		extern int chan_eth_info(an_ifinfo_t i[], int *c);

		count = NODEOS_MAXIFS - numifs;
		err = chan_eth_info(infop, &count);
		if (err == 0) {
			numifs += count;
			infop += count;
		}
	}
#endif

	/*
	 * Assign logical IDs
	 */
	for (i = 0; i < numifs; i++)
		ifinfo[i].id = i;
}

an_ifinfo_t *
if_getinfo(int *count)
{
	int num;

	assert(count != 0);

	num = *count;
	if (num == 0)
		num = numifs;
	else if (num > numifs)
		num = numifs;
	*count = num;

	return ifinfo;
}

#ifdef CHAN_IF
/*
 * Hack "if" channel type support.
 *
 * "if" channels are "generic" network interface channels.  They return
 * packet payloads stripped of any interface headers; e.g., ethernet packets
 * stripped of the 14 byte ethernet header.  Mapping of physical interfaces
 * to logical "if" names is done via the ID contained in the ifinfo struct.
 */

void
chan_if_init(void)
{
	if_setup();
}

/*
 * Translate "if" name to underlying interface(s)
 */
static int
chan_if_find(char *name, an_ifinfo_t **infop, an_iftype_t *iftype)
{
	an_ifinfo_t *info;
	int id;

#if 0
	if (strcmp(name, "if") == 0) {
		*iftype = ANIF_ANY;
		*infop = 0;
		return 0;
	}
	if (strcmp(name, "ifeth") == 0) {
		*iftype = ANIF_ETH;
		*infop = 0;
		return 0;
	}
#endif

	id = atoi(&name[2]);
	if (id < 0 || id >= numifs)
		return EINVAL;

	for (info = ifinfo; info < &ifinfo[numifs]; info++)
		if (id == info->id)
			break;
	if (info == &ifinfo[numifs])
		return EINVAL;

	*infop = info;
	*iftype = info->type;
	return 0;
}

int
chan_if_create(ani_chan_t *chan)
{
	an_iftype_t iftype;
	an_ifinfo_t *info;
	int err;

	err = chan_if_find(ani_chan_name(chan), &info, &iftype);
	if (err)
		return err;

	/*
	 * XXX support multi-device channels some day
	 */
	if (info == 0)
		return ENODEV;

	switch (iftype) {
	case ANIF_ETH:
#ifdef CHAN_ETH
	{
		chan->flags &= ~CHAN_ISRAW;

		/*
		 * XXX hackery.  Replace user-specified proto spec
		 * with one crafted to reflect the actual interface.
		 */
		strcpy(chan->proto.data, info->pname);
		chan->proto.len = strlen(info->pname);

		err = chan_eth_create(chan);
		break;
	}
#endif
	case ANIF_ANY:
	case ANIF_ATM:
		panic("create: Bad IF type");
	}

	return err;
}

void
chan_if_destroy(ani_chan_t *chan)
{
	an_iftype_t iftype;
	an_ifinfo_t *info;
	int err;

	err = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(err == 0);
	assert(info != 0);

	switch (iftype) {
	case ANIF_ETH:
#ifdef CHAN_ETH
		chan_eth_destroy(chan);
		break;
#endif
	case ANIF_ANY:
	case ANIF_ATM:
		panic("destroy: Bad IF type");
	}
}

int
chan_if_enable(ani_chan_t *chan)
{
	an_iftype_t iftype;
	an_ifinfo_t *info;
	int err;

	err = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(err == 0);
	assert(info != 0);

	switch (iftype) {
	case ANIF_ETH:
#ifdef CHAN_ETH
		err = chan_eth_enable(chan);
		break;
#endif
	case ANIF_ANY:
	case ANIF_ATM:
		panic("enable: Bad IF type");
	}

	return err;
}

void
chan_if_disable(ani_chan_t *chan)
{
	an_iftype_t iftype;
	an_ifinfo_t *info;
	int err;

	err = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(err == 0);
	assert(info != 0);

	switch (iftype) {
	case ANIF_ETH:
#ifdef CHAN_ETH
		chan_eth_disable(chan);
		break;
#endif
	case ANIF_ANY:
	case ANIF_ATM:
		panic("disable: Bad IF type");
	}
}

void
chan_if_send(ani_chan_t *chan, ani_pbuf_t *pbuf)
{
	an_iftype_t iftype;
	an_ifinfo_t *info;
	int err;

	err = chan_if_find(ani_chan_name(chan), &info, &iftype);
	assert(err == 0);
	assert(info != 0);

	switch (iftype) {
	case ANIF_ETH:
#ifdef CHAN_ETH
		chan_eth_send(chan, pbuf);
		break;
#endif
	case ANIF_ANY:
	case ANIF_ATM:
		panic("send: Bad IF type");
	}
}

#endif
