/*
 * Copyright (c) 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>
#include <sys/types.h>
#ifdef OSKIT
#include <oskit/io/bufio.h>
#endif

#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include <nodeos/impl/ani_pbuf.h>

/*
 * Packet buffer manipulation
 */

/*
 * Allocate system-wide packet buffer memory.
 * Right now we just malloc every time.  Someday we might manage
 * a cache of memory.
 *
 * @allocation: "size" bytes charged to the given flow.
 */
static void *
pbuffer_memalloc(ani_flow_t *flow, size_t size)
{
	assert_ani_flow_valid(flow);
	assert(size > 0);

	/* XXX round up size or verify it is a valid size */

	return ani_mempool_alloc(ani_flow_mempool(flow), size);
}

/*
 * Free system-wide packet buffer memory.
 * Called with a pbuf when its reference count drops to zero.
 * Frees just the buffer, not the header.
 *
 * @deallocation: pbuf->bufsize bytes returned to pbuffer's flow.
 */
static void
pbuffer_memfree(ani_pbuf_t *pbuf, void *arg)
{
	assert(pbuf != 0);
	assert_ani_flow_valid(pbuf->flow);
	assert(pbuf->bufdata != 0);
	assert(pbuf->bufsize > 0);

	ani_mempool_free(ani_flow_mempool(pbuf->flow),
			 pbuf->bufdata, pbuf->bufsize);
	pbuf->bufdata = 0;
}

/*
 * Initialize the given pbuf with the indicated info.
 */
void
ani_pbuf_init_external(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan,
		       void *bdata, size_t bsize, size_t off, size_t size,
		       ani_pbuf_free_t ffunc, void *farg)
{
	assert(pbuf != 0);
	assert_ani_flow_valid(flow);
	assert(bdata != 0);
	assert(off >= 0);
	assert(off + size <= bsize);
	assert(ffunc != 0);

	pbuf->bufdata = bdata;
	pbuf->bufsize = bsize;
	pbuf->refcount = 1;
	pbuf->off = off;
	pbuf->size = size;
	pbuf->ffunc = ffunc;
	pbuf->farg = farg;
	pbuf->flow = flow;
	pbuf->chan = chan;
}

/*
 * Allocate a buffer and associate it with the given pbuf object.
 */
int
ani_pbuf_init(ani_pbuf_t *pbuf, ani_flow_t *flow, ani_chan_t *chan, size_t bsize,
	      size_t off, size_t size)
{
	void *bdata;

	bdata = pbuffer_memalloc(flow, bsize);
	if (bdata == 0) {
		ani_mempool_free(ani_flow_mempool(flow), pbuf, sizeof *pbuf);
		return ENOMEM;
	}

	ani_pbuf_init_external(pbuf, flow, chan, bdata, bsize, off, size,
			       pbuffer_memfree, 0);

	return 0;
}

void
ani_pbuf_deinit(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);

	if (pbuf->bufdata != 0)
		ani_pbuf_release(pbuf);
}

/*
 * Set the offset of the active region of the pbuffer.
 * Implicitly changes the size of the active region of the buffer.
 */
int
ani_pbuf_setoffset(ani_pbuf_t *pbuf, size_t newoff, void **out_data)
{
	int eoffset;

	assert(pbuf != 0);
	assert(pbuf->bufdata != 0);

	eoffset = pbuf->off + pbuf->size;
	if (newoff < 0 || newoff > eoffset)
		return EINVAL;

	pbuf->off = newoff;
	pbuf->size = eoffset - newoff;
	if (out_data)
		*out_data = pbuf->bufdata + pbuf->off;

	return 0;
}

/*
 * Set the size of the active region of the pbuffer.
 */
int
ani_pbuf_setsize(ani_pbuf_t *pbuf, size_t newsize, void **out_data)
{
	assert(pbuf != 0);
	assert(pbuf->bufdata != 0);

	if (newsize < 0 || pbuf->off + newsize > pbuf->bufsize)
		return EINVAL;

	pbuf->size = newsize;
	if (out_data)
		*out_data = pbuf->bufdata + pbuf->off;

	return 0;
}

/*
 * Open a buffer of to its max size.
 * Returns a pointer to the active region.
 */
void *
ani_pbuf_setmax(ani_pbuf_t *pbuf)
{
	assert(pbuf != 0);
	assert(pbuf->bufdata != 0);

	pbuf->off = 0;
	pbuf->size = pbuf->bufsize;

	return pbuf->bufdata;
}

/*
 * Swap the memory buffers of two buffer objects.
 */
void
ani_pbuf_swap(ani_pbuf_t *pbuf1, ani_pbuf_t *pbuf2)
{
	ani_pbuf_t tmp;

	assert(pbuf1->refcount == 1);
	assert(pbuf2->refcount == 1);

	tmp = *pbuf1;
	*pbuf1 = *pbuf2;

	/*
	 * Retain flow and channel association
	 * XXX need to adjust accounting if buffer sizes don't match.
	 */
	pbuf1->flow = tmp.flow;
	pbuf1->chan = tmp.chan;
	tmp.flow = pbuf2->flow;
	tmp.chan = pbuf2->chan;

	*pbuf2 = tmp;
}

#ifdef OSKIT
static void
ani_pbuf_bufiofree(ani_pbuf_t *pbuf, void *arg)
{
	oskit_bufio_t *bufio = arg;

	assert(bufio != 0);
	assert(pbuf->bufdata != 0);

	oskit_bufio_unmap(bufio, pbuf->bufdata, 0, pbuf->bufsize);
	oskit_bufio_release(bufio);
	pbuf->bufdata = 0;
}

/*
 * Wrap an OSKit bufio with a NodeOS packet buffer.
 * The caller should have a reference to the bufio.  This routine will
 * take a bufio reference if successful.
 */
int
ani_pbuf_frombufio(ani_flow_t *flow, ani_chan_t *chan, oskit_bufio_t *bufio,
		   size_t off, size_t size, ani_pbuf_t **out_pbuf)
{
	int err;
	void *bdata;
	size_t bsize;
	oskit_off_t bs;
	ani_pbuf_t *pbuf;

	assert_ani_flow_valid(flow);
	assert(bufio != 0);
	assert(off >= 0);

	/*
	 * Determine the size of the incoming packet and map it
	 */
	err = oskit_bufio_getsize(bufio, &bs);
	if (err != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return err;
	}
	bsize = (size_t)bs;
	if (off + size > bsize) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return EINVAL;
	}
	err = oskit_bufio_map(bufio, &bdata, 0, bsize);
	if (err != 0) {
		CHANSTATS(chan->stats.droppackets++);
		CHANSTATS(chan->stats.dropfailedbuf++);
		return err;
	}

	/*
	 * Get a large enough buffer from the channel.
	 * If it can't pay, we drop the packet.
	 */
	err = ani_chan_pbuf_alloc(chan, bsize, &pbuf);
	if (err) {
		oskit_bufio_unmap(bufio, bdata, 0, bsize);
		return err;
	}
	assert(pbuf->flow == flow);

#ifndef OSKIT_SKBUFF_HACKS
	/*
	 * Right now we must copy all but max sized packets.  The danger is
	 * that a device will hand us back a packet in a buffer that is less
	 * than our BSIZE, and that buffer would get passed back to the user
	 * and eventually recycled and handed back to the device.
	 * This happens because we cannot directly control memory allocation
	 * done by the linux ethernet device drivers.  Some drivers such
	 * as the eepro100 will copy small packets into custom-sized skbuffs
	 * so we cannot guarentee fixed-size buffers within the nodeos
	 * unless we always copy them here.  In the future we will customize
	 * some OSKit interfaces to allow more direct control of the memory
	 * allocation of skbuffs.
	 */
	if (bsize < 1514) {
		memcpy(ani_pbuf_data(pbuf), bdata+off, size);
		CHANSTATS(chan->stats.bufcopies++);
	}
#else
	/*
	 * We copy small packets for two reasons, 1) because its possibly
	 * more efficient, 2) because we don't want to be swapping buffers
	 * that are too small into our packet buffer pool.
	 */
	if (bsize <= NODEOS_CHAN_ETH_CSIZE) {
		memcpy(ani_pbuf_data(pbuf), bdata+off, size);
		CHANSTATS(chan->stats.bufcopies++);
	}
#endif
	/*
	 * Packet is large enough to swap
	 */
	else {
		/*
		 * XXX free the actual data buffer as we cannot return it
		 * to the OSKit driver to use right now.
		 */
		pbuf->ffunc(pbuf, pbuf->farg);
		/*
		 * Intialize the pbuf with the new buffer info.
		 */
		ani_pbuf_init_external(pbuf, flow, chan, bdata, bsize, off,
				       size, ani_pbuf_bufiofree, bufio);
		/*
		 * Keep the bufio alive
		 */
		oskit_bufio_addref(bufio);
	}

	*out_pbuf = pbuf;
	return 0;
}

int
ani_pbuf_tobufio(ani_pbuf_t *pbuf, void *hdr, size_t hsize,
		 oskit_bufio_t **out_bufio)
{
	oskit_bufio_t *bufio;
	oskit_size_t psize;
	void *bdata;
	int err;

	/*
	 * Packet buffer is already a wrapped bufio,
	 *
	 * We can just use the existing bufio if, after pre-pending any
	 * header, the pbuf offset would be zero.  Note that we do not
	 * actually modify the pbuf.
	 */
	if (pbuf->ffunc == ani_pbuf_bufiofree && pbuf->off - hsize == 0) {
		if (hsize != 0)
			memcpy(pbuf->bufdata, hdr, hsize);
		bufio = pbuf->farg;
		oskit_bufio_addref(bufio);
		*out_bufio = bufio;
		return 0;
	}

	/*
	 * Otherwise cons up a bufio and copy the header/data.
	 *
	 * XXX should request a bufio from the target netio implemention
	 * (e.g., the linux device drivers).
	 */
	psize = ani_pbuf_size(pbuf);
	bufio = oskit_bufio_create(hsize + psize);
	if (bufio == 0)
		return ENOMEM;

	/*
	 * Copy the header/data.
	 */
	err = oskit_bufio_map(bufio, &bdata, 0, hsize + psize);
	if (err == 0) {
		if (hsize)
			memcpy(bdata, hdr, hsize);
		memcpy(bdata+hsize, ani_pbuf_data(pbuf), psize);
		oskit_bufio_unmap(bufio, bdata, 0, hsize + psize);
	} else {
		oskit_size_t off = 0;

		if (hsize) {
			err = oskit_bufio_write(bufio, hdr, 0, hsize, &off);
			if (err != 0 || off != hsize) {
				oskit_bufio_release(bufio);
				return err;
			}
		}
		err = oskit_bufio_write(bufio, ani_pbuf_data(pbuf), off,
					psize, &off);
		if (err != 0 || off != psize) {
			oskit_bufio_release(bufio);
			return err;
		}
	}

	*out_bufio = bufio;
	return 0;
}

#endif
