/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <assert.h>

#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_mutex.h>
#include <nodeos/impl/ani_cond.h>
#include <nodeos/impl/ani_thread.h>

/*
 * N.B. NodeOS mutexes are not cancelable.
 * This would obviously have to change if the NodeOS doesn't trust the EE.
 */

int
ani_mutex_init(ani_mutex_t *newmutex)
{
	assert(newmutex != 0);
	ani_ilock_init(&newmutex->ilock);
	return 0;
}

/* ani_mutex_lock is in <nodeos/impl/ani_mutex.h> */
/* ani_mutex_trylock is in <nodeos/impl/ani_mutex.h> */
/* ani_mutex_unlock is in <nodeos/impl/ani_mutex.h> */

int
ani_mutex_deinit(ani_mutex_t *mutex)
{
	ani_ilock_deinit(&mutex->ilock);
	return 0;
}


int
ani_cond_init(ani_cond_t *newcond)
{
	assert(newcond != 0);
	ani_icond_init(&(newcond->icond));
	ani_ilock_init(&(newcond->ilock));
	return 0;
}

void
ani_cond_timedwait(ani_cond_t *cond, ani_mutex_t *mutex,
		   an_timespec_t *abstime)
{
	ani_cond_lock(cond);

	/*
	 * Release the mutex and wait.
	 * Note that we are still atomic here since no one can signal the
	 * condition until we release the cond lock.
	 */
	ani_mutex_unlock(mutex);
	ani_thread_condtimedwait(cond, abstime);

	/*
	 * Now unlock the condition before attempting to
	 * reacquire the mutex.  There is no race here.
	 */
	ani_cond_unlock(cond);
	
	ani_mutex_lock(mutex);
}

void
ani_cond_signal(ani_cond_t *cond)
{
	/*
	 * Use the cond's internal lock to synchronize correctly with
	 * waiters.
	 */

	ani_cond_lock(cond);
	ani_icond_signal(&(cond->icond));
	ani_cond_unlock(cond);
}

void
ani_cond_broadcast(ani_cond_t *cond)
{
	ani_cond_lock(cond);
	ani_icond_broadcast(&(cond->icond));
	ani_cond_unlock(cond);
}

int
ani_cond_deinit(ani_cond_t *cond)
{
	ani_ilock_deinit(&(cond->ilock));
	ani_icond_deinit(&(cond->icond));
	return 0;
}
