/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <pthread.h>
#include <sched.h>

#include <nodeos/an_thread.h>
#include <nodeos/impl/an_error.h>
#include <nodeos/impl/ani_thread.h>

#include "os.h"
#include <nodeos/impl/ani_flow.h>
#include <nodeos/impl/ani_chan.h>
#include <nodeos/impl/ani_mempool.h>
#include "threadpool.h"
#include "misc.h"

//#define VERBOSE

#ifdef VERBOSE
#define VPRINTF(args...) printf(args)
#else
#define VPRINTF(args...)
#endif


static pthread_key_t nodeos_thread;	/* thread struct for this pthread */
static pthread_key_t nodeos_threaddata;	/* nodeos exported thread data */

/*
 * For hack thread_sleep implementation.
 * Cond is never signalled, just allows us to do a timed wait.
 */
static pthread_mutex_t sleepmutex;
static pthread_cond_t sleepcond;

static void ani_thread_done(void *arg);
static void ani_thread_cleanup(void *arg);
static void *ani_thread_loop(void *arg);
static void ani_thread_intr_init(void);

void
ani_thread_setup(void)
{
	int err;

	/*
	 * Establish a thread_specific data key to hold the ani_thread_t
	 * associated with a pthread.  This is used to determine the
	 * current thread.  Init to zero for this primortial thread.
	 */
	err = pthread_key_create(&nodeos_thread, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_thread, 0);
	assert(err == 0);

	/*
	 * Also need a key to use for the exported ani_thread_setdata
	 * interface.
	 */
	err = pthread_key_create(&nodeos_threaddata, 0);
	assert(err == 0);
	err = pthread_setspecific(nodeos_threaddata, 0);
	assert(err == 0);

	/*
	 * For thread sleep
	 */
	err = pthread_mutex_init(&sleepmutex, 0);
	assert(err == 0);
	err = pthread_cond_init(&sleepcond, 0);
	assert(err == 0);

	ani_thread_intr_init();
}

ani_thread_t *
ani_thread_current(void)
{
	/*
	 * In the current implementation has no control over scheduling
	 * of threads, so we have to query the pthread state.
	 */
	return pthread_getspecific(nodeos_thread);
}

/*
 * Called during thread pool initialization to ensure that thread parameters
 * are compatible with the underlying thread implementation.
 */
int
ani_thread_checkattrs(an_cpuspec_t *cspec)
{
	pthread_attr_t attr;
	int err;

	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, cspec->stacksize);
	return err;
}

int
ani_thread_spawn(threadpool_t *tpool, ani_thread_t **out_thread)
{
	ani_thread_t *thread;
	pthread_attr_t attr;
	int err;

	assert(threadpool_islocked(tpool));

	if (tpool->curthreads >= tpool->params.max)
		return EAGAIN;

	thread = ani_mempool_alloc(ani_flow_mempool(tpool->flow),
				   sizeof *thread);
	if (thread == 0)
		return ENOMEM;
	memset(thread, 0, sizeof *thread);

	/*
	 * Set up the attributes for the thread.
	 * Currently we only set the stack size.
	 * (XXX and global scope for the benefit of Solaris)
	 */
	err = pthread_attr_init(&attr);
	assert(err == 0);
	err = pthread_attr_setscope(&attr, PTHREAD_SCOPE_SYSTEM);
	assert(err == 0);
	err = pthread_attr_setstacksize(&attr, tpool->params.stacksize);
	assert(err == 0);

	/*
	 * Create the pthread, it will block immediately on the pool lock.
	 */
	thread->pool = tpool;
	err = pthread_create(&thread->id, &attr, ani_thread_loop, thread);
	if (err) {
		ani_mempool_free(ani_flow_mempool(tpool->flow),
				 thread, sizeof *thread);
		return err;
	}
	err = pthread_detach(thread->id);
	assert(err == 0);

	/* Set the default scheduling priority */
	thread->nprio = AN_THREAD_PRIO_NORMAL;
	err = os_sched_setprio(thread, thread->nprio);
	assert(err == 0);

	/* We are committed, finish up */
	ani_ilock_init(&thread->lock);
	thread->flags = THREAD_INKERNEL;
	thread->pool = tpool;
	ani_icond_init(&thread->waiting);
	ani_icond_init(&thread->waitedon);
	thread->next = tpool->threads;
	tpool->threads = thread;
	tpool->curthreads++;
	threadpool_reference(tpool);

	if (out_thread)
		*out_thread = thread;

	return 0;
}

/*
 * All NodeOS threads spend their lives here with brief forays into EE/AA code.
 */
static void *
ani_thread_loop(void *arg)
{
	ani_thread_t *thread = arg;
	threadpool_t *tpool;
	struct threadwork *w;
	struct threadwork wtmp;

	/*
	 * Record our NodeOS identity in the pthread so thread_current works.
	 */
	assert(thread != 0);
	pthread_setspecific(nodeos_thread, thread);

	/*
	 * Get the threadpool lock.
	 * We will block here until our creator is finished initializing us.
	 */
	tpool = thread->pool;
	assert(tpool != 0);
	threadpool_lock(tpool);
	VPRINTF("TP%p: thread %p started\n", tpool, thread);

	/*
	 * Now that we are official, push a cleanup handler, and go to work.
	 */
	pthread_cleanup_push(ani_thread_cleanup, thread);
	while (1) {
		/*
		 * If there is no work to do we either block waiting for
		 * more or die if there are enough threads in reserve.
		 */
		while (thread->work == 0 && tpool->workhead == 0) {
			if (tpool->freethreads >= tpool->reserve)
				goto die;
			thread->freenext = tpool->freelist;
			tpool->freelist = thread;
			tpool->freethreads++;
#ifdef THREADPOOL_STATS
			if (tpool->freethreads > tpool->stats.maxcached)
				tpool->stats.maxcached = tpool->freethreads;
#endif
			if (tpool->flags & TPOOL_WAITING) {
				tpool->flags &= ~TPOOL_WAITING;
				ani_icond_broadcast(&tpool->waiting);
			}
			VPRINTF("TP%p: thread %p sleeping, freethreads=%d\n",
				tpool, thread, tpool->freethreads);
			ani_icond_wait(&thread->waiting, &tpool->lock);
			VPRINTF("TP%p: thread %p awake, freethreads=%d\n",
				tpool, thread, tpool->freethreads);
		}
		assert(thread->freenext == 0);

#ifdef THREADPOOL_STATS
		{
			int active = tpool->curthreads - tpool->freethreads;

			if (active > tpool->stats.maxactive)
				tpool->stats.maxactive = active;
		}
#endif

		/*
		 * First look for work specifically targetted to this thread.
		 * Otherwise, grab the first item off the list.
		 */
		w = thread->work;
		if (w == 0) {
			w = tpool->workhead;
			if (w->next == 0) {
				assert(tpool->worktail == w);
				tpool->workhead = tpool->worktail = 0;
			} else
				tpool->workhead = w->next;
			tpool->worklen--;
		} else
			thread->work = 0;

		/*
		 * Make a local copy so we can free the heap element
		 */
		wtmp = *w;
		ani_mempool_free(ani_flow_mempool(tpool->flow), w, sizeof *w);
		w = &wtmp;

		thread->chan = w->chan;

		VPRINTF("TP%p: thread %p doing %p(%p)\n",
			tpool, thread, w->func, w->farg);

		/*
		 * Run the desired function.
		 *
		 * The sigsetjmp catches threads that call ani_thread_exit
		 * explicitly (as opposed to returning from the registered
		 * upcall func).  Those threads will siglongjmp back to
		 * here unwinding their stacks.
		 *
		 * The cleanup handler ensure that anybody awaiting this
		 * thread is properly awakened in the event of a cancellation.
		 * Note the "1" argument to cleanup_pop that causes
		 * ani_thread_done to be called even when no cancellation
		 * occured.
		 */
		pthread_cleanup_push(ani_thread_done, thread);
		threadpool_unlock(tpool);
		if (sigsetjmp(thread->jmpbuf, 0) == 0) {
			if (w->inkern == 0) {
				thread->flags |= THREAD_INTSON;
				NODEOS_LEAVE;
				w->func(w->farg[0], w->farg[1]);
				NODEOS_ENTER;
			} else
				w->func(w->farg[0], w->farg[1]);
		} else
			thread = ani_thread_current();
		tpool = thread->pool;
		pthread_cleanup_pop(1);	/* calls ani_thread_done */
		threadpool_lock(tpool);
		VPRINTF("TP%p: thread %p did %p(%p)\n",
			tpool, thread, w->func, w->farg);
	}
 die:
	VPRINTF("TP%p: thread %p dying, freethreads=%d\n",
		tpool, thread, tpool->freethreads);
	threadpool_unlock(thread->pool);

	/*
	 * Call pthread_exit explicitly here instead of returning
	 * since the cleanup handler must be "in scope" when we exit.
	 */
	pthread_exit(0);
	pthread_cleanup_pop(1);

	return 0;
}

void
ani_thread_kill(ani_thread_t *thread)
{
	int err;

	assert(thread != ani_thread_current());

	err = pthread_cancel(thread->id);
	assert(err == 0);
}

void
ani_thread_exit(void)
{
	/* Just exit, the actual work is done in the cleanup handler */
	pthread_exit(0);
}

/*
 * Wake any thread waiting for this one to finish and wait for it to
 * acknowledge.  Called from the main thread loop (or pthread_exit) to
 * ensure a thread's previous commitments are completed before returning
 * to the common pool (or exiting).
 */
static void
ani_thread_done(void *arg)
{
	ani_thread_t *thread = arg;
	threadpool_t *tpool = thread->pool;

	if (thread->chan) {
		ani_chan_thread_exit(thread->chan);
		thread->chan = 0;
	}

	threadpool_lock(tpool);
	if (thread->flags & THREAD_WAITEDON) {
		VPRINTF("TP%p: thread %p finishing\n", tpool, thread);
		thread->flags &= ~THREAD_WAITEDON;
		ani_icond_signal(&thread->waitedon);
		thread->flags |= THREAD_DONE;
		while ((thread->flags & THREAD_DONE) != 0)
			ani_icond_wait(&thread->waiting, &tpool->lock);
		VPRINTF("TP%p: thread %p finished\n", tpool, thread);
	}
	assert((thread->flags & (THREAD_WAITEDON|THREAD_DONE)) == 0);
	threadpool_unlock(tpool);
}

/*
 * Cleanup a thread's resources.
 * This is the pthread cleanup handler called for every thread when it
 * exits or is canceled.
 */
static void
ani_thread_cleanup(void *arg)
{
	ani_thread_t *thread;
	ani_thread_t **pth, *th;
	threadpool_t *tpool;

	thread = arg;
	assert(thread == ani_thread_current());
	assert((thread->flags & (THREAD_WAITEDON|THREAD_DONE)) == 0);
	tpool = thread->pool;
	assert_threadpool_valid(tpool);

	threadpool_lock(tpool);
	VPRINTF("TP%p: cleanup thread %p\n", tpool, thread);
	for (pth = &tpool->threads; (th = *pth) != 0; pth = &th->next)
		if (th == thread)
			break;
	assert(th != 0);
	*pth = thread->next;

	tpool->curthreads--;

	ani_ilock_deinit(&thread->lock);
	ani_icond_deinit(&thread->waitedon);
	ani_mempool_free(ani_flow_mempool(tpool->flow), thread, sizeof *thread);
	threadpool_unlock(tpool);
	threadpool_release(tpool);
}

void
ani_thread_yield(void)
{
	sched_yield();
}

/*
 * XXX hack, uninterruptable (though cancelable) sleep implementation.
 * XXX assumes that an_timespec is same as timespec
 *
 * I'm not convinced ani_thread_sleep is all that useful or I would spend some
 * time doing it right.
 */
void
ani_thread_sleep(an_timespec_t *delay)
{
	int err;
	an_timespec_t wakeup;

	time_future(&wakeup, delay);

	err = pthread_mutex_lock(&sleepmutex);
	assert(err == 0);

	do {
		pthread_cleanup_push((void (*)(void *))pthread_mutex_unlock,
				     &sleepmutex);
		err = pthread_cond_timedwait(&sleepcond, &sleepmutex, &wakeup);
		pthread_cleanup_pop(0);
	} while (err == 0);
	assert(err == EINVAL || err == ETIMEDOUT);

	err = pthread_mutex_unlock(&sleepmutex);
	assert(err == 0);
}

void *
ani_thread_getdata(void)
{
	return pthread_getspecific(nodeos_threaddata);
}

void
ani_thread_setdata(void *data)
{
	int err;

	err = pthread_setspecific(nodeos_threaddata, data);
	assert(err == 0);
}

int
ani_thread_getprio(ani_thread_t *thread)
{
	return thread->nprio;
}

int
ani_thread_setprio(ani_thread_t *thread, int nprio)
{
	int err;

	err = os_sched_setprio(thread, nprio);
	if (err == 0)
		thread->nprio = nprio;

	return err;
}

/*
 * Get various bits of stack information on a specific
 * thread.  The thread may be the current thread.
 * If the target thread is running (on an SMP machine)
 * the results are undefined.
 *
 * XXX this call may evolve into a more generic ani_thread_info call.
 */
int
ani_thread_stackinfo(ani_thread_t *thread, void **base, size_t *size, void **ptr)
{
	int err;
#ifdef OSKIT
	struct pthread_state ps;

	err = oskit_pthread_getstate(thread->id, &ps);
	if (err == 0) {
		*base = (void *)ps.stackbase;
		*size = ps.stacksize;
		*ptr = (void *)ps.stackptr;
	}
#else
	err = ENODEV;
#endif
	return err;
}

/*
 * Thread "interrupt" support.
 *
 * With pthreads, signals are the only way we can do this.
 * Pthread cancelation implies the death of the canceled thread,
 * NodeOS interrupts do not.
 */
#include <signal.h>

#define SIG_THREADINTR	SIGUSR1 /* XXX */
sigset_t ani_thread_intrmask;

static void ani_thread_intr_handler(int sig, siginfo_t *si, void *sc);

static void
ani_thread_intr_init(void)
{
	struct sigaction act;

	sigemptyset(&ani_thread_intrmask);
	sigaddset(&ani_thread_intrmask, SIG_THREADINTR);

	/*
	 * Catch relevant signals.
	 */
#ifdef SA_SIGINFO
	act.sa_flags = SA_SIGINFO;
	act.sa_sigaction = ani_thread_intr_handler;
#else
	act.sa_handler = ani_thread_intr_handler;
#endif
	sigemptyset(&act.sa_mask);

	sigaction(SIG_THREADINTR, &act, 0);
	sigaction(SIGILL, &act, 0);
#ifdef SIGEMT
	sigaction(SIGEMT, &act, 0);
#endif
	sigaction(SIGFPE, &act, 0);
	sigaction(SIGBUS, &act, 0);
	sigaction(SIGSEGV, &act, 0);
}

void
ani_thread_interrupt(ani_thread_t *thread)
{
	int err;
	ani_cond_t *cond;

	ani_thread_lock(thread);

	/*
	 * Send a signal to said thread
	 */
	err = pthread_kill(thread->id, SIG_THREADINTR);
	assert(err == 0);

	/*
	 * If thread is in a cond wait, wake it up.
	 * This gets really strange because we must acquire the cond
	 * lock before the thread lock (because of how ani_thread_condwait
	 * works).
	 */
 again:
	cond = thread->cond;
	if (cond) {
		/*
		 * Cond lock is held.  Release the thread lock and wait
		 * for the cond.  Then reaquire the thread lock and see
		 * if it is still waiting on the same cond.  If not,
		 * start over.
		 */
		if (ani_cond_trylock(cond)) {
			ani_thread_unlock(thread);
			ani_cond_lock(cond);
			ani_thread_lock(thread);
			if (thread->cond != cond) {
				ani_cond_unlock(cond);
				goto again;
			}
		}
		ani_icond_broadcast(&cond->icond);
		ani_cond_unlock(cond);
	}

	ani_thread_unlock(thread);
}

void
ani_thread_block_interrupt(void)
{
	ani_thread_current()->flags &= ~THREAD_INTSON;
}

void
ani_thread_unblock_interrupt(void)
{
	ani_thread_current()->flags |= THREAD_INTSON;
}

void
ani_thread_sethandler(an_threadintr_func_t handler)
{
	ani_thread_current()->handler = handler;
}

/*
 * Signal handler for all exceptions and NodeOS thread interrupts
 */
static void
ani_thread_intr_handler(int sig, siginfo_t *si, void *sc)
{
	ani_thread_t *thread = ani_thread_current();
	int intson, type;

	if (thread == 0 || (thread->flags & THREAD_INKERNEL)) {
		os_dump_state(sc);
		panic("NodeOS: got signal %d\n", sig);
	}

	NODEOS_ENTER;

	/*
	 * No handler for thread, just exit
	 *
	 * XXX should kill the flow.  Rationale: if an EE plans on
	 * interrupting a thread, it should be prepared to catch it.
	 * And if a thread faults, that is clearly a flow/EE error.
	 * By just killing the thread as we do now, we have effectively
	 * introduced an unsafe ani_thread_kill operation.  Me bad.
	 */
	 if (thread->handler == 0) {
		 printf("ani_thread_intr_handler: thread has no handler:\n");
		 os_dump_state(sc);
		 pthread_cancel(thread->id);
		 return;
	 }

	/*
	 * Otherwise, call the handler with interrupts disabled
	 */
	intson = (thread->flags & THREAD_INTSON);
	thread->flags &= ~THREAD_INTSON;

	/*
	 * Map the signal type to the appropriate AN_THREAD_ constant.
	 */
	switch (sig)
	{
	case SIG_THREADINTR:
		assert(intson);
		type = AN_THREAD_INTR;
		break;
	case SIGFPE:
		type = AN_THREAD_TRAP_FPU;
		break;
#if defined(SIGBUS)
	case SIGBUS:  /* map to SEGV */
#endif
	case SIGSEGV:
	default:
		type = AN_THREAD_TRAP_SEGV;
		break;
	}

	NODEOS_LEAVE;
	thread->handler(type, sc);
	NODEOS_ENTER;

	/* XXX NODEOS_LEAVE sans signal code */
	thread->flags |= intson;
	thread->flags &= ~THREAD_INKERNEL;

	NODEOS_LEAVE;
}

/*
 * Block a thread on the specified condition variable.
 * The thread may be awakened either by the condition being signaled through
 * an API call or because of ani_thread_interrupt.  We do not differentiate the
 * two cases here.
 *
 * If abstime is not NULL, then the cond wait may timeout if the given
 * absolute time is reached.
 *
 * Called with thread and condition locked.
 */
void
ani_thread_condtimedwait(ani_cond_t *cond, an_timespec_t *abstime)
{
	ani_thread_t *thread = ani_thread_current();
	assert(thread->cond == NULL);

	thread->cond = cond;
	if (abstime == NULL) 
	{
		ani_icond_wait(&cond->icond, &cond->ilock);
	}
	else
	{
		ani_icond_timedwait(&cond->icond, &cond->ilock,
				    abstime->tv_sec, abstime->tv_nsec);
	}
	
	thread->cond = NULL;
}
