/*
 * Copyright (c) 1999, 2000 The University of Utah and the Flux Group.
 * All rights reserved.
 * This file is part of Janos.
 * 
 * Janos is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License, version 2, as
 * published by the Free Software Foundation.
 * 
 * Janos is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details, in the file
 * COPYING.  If that file is not present, write to the Free
 * Software Foundation, 59 Temple Place - Suite 330, Boston, MA  
 * 02111-1307, USA
 * 
 * The University of Utah and the Flux Group requests users to
 * return to flux-dist@cs.utah.edu any improvements that they make.
 */

#ifndef _THREADPOOL_H_
#define _THREADPOOL_H_

#undef THREADPOOL_STATS

#include <nodeos/impl/ani_ilock.h>
#include <nodeos/impl/ani_resource.h>
#include <nodeos/impl/ani_thread.h>

/*
 * The threadpool: a group of threads sharing the same scheduling params
 */

struct ani_flow;
struct ani_chan;

typedef void (*thread_func_t)(void *, void *);

struct threadwork {
	struct threadwork *next;/* next item on list */
	int inkern;		/* function should run in kernel context */
	struct ani_chan *chan;	/* channel to associate thread with */
	thread_func_t func;	/* function the thread should start in */
	void *farg[2];		/* and arguments for that function */
};

struct threadpool {
	struct ani_flow *flow;		/* associated flow */
	struct ani_chan *chan;		/* associated channel */
	int flags;			/* various flags */
	int curthreads;			/* current # of threads in pool */
	ani_thread_t *threads;		/* list of current threads */
	int reserve;			/* desired number of free threads */
	int freethreads;		/* current # of free threads */
	ani_thread_t *freelist;		/* list of free threads */
	ani_icond_t waiting;		/* place to wait for this pool */
	int worklen;			/* length of work queue */
	struct threadwork *workhead;	/* head of work queue */
	struct threadwork *worktail;	/* tail of work queue */
	an_cpuspec_t params;		/* resource parameters */
	int refs;			/* reference count */
	ani_ilock_t lock;		/* lock for entire pool */
#ifdef THREADPOOL_STATS
	struct {
		int starts;
		int creates;
		int queued;
		int failed;
		int maxactive;
		int maxcached;
		int maxqueued;
	} stats;
#endif
};
typedef struct threadpool threadpool_t;

#define TPOOL_DYING	1
#define TPOOL_WAITING	2

#define threadpool_lock(t)	ani_ilock_lock(&(t)->lock)
#define threadpool_unlock(t)	ani_ilock_unlock(&(t)->lock)
#define threadpool_islocked(t)	ani_ilock_islocked(&(t)->lock)

#define threadpool_reference(t)	((t)->refs++)
#define threadpool_release(t)	if (--(t)->refs == 0) threadpool_cleanup(t)

int threadpool_create(struct ani_flow *flow, struct ani_chan *chan,
		      an_cpuspec_t *cspec, threadpool_t **out_pool);
void threadpool_kill(threadpool_t *tpool, struct ani_chan *chan);
void threadpool_destroy(threadpool_t *tpool);
void threadpool_cleanup(threadpool_t *tpool);

#ifdef MEMDEBUG
#define assert_threadpool_valid(t) \
	assert((t) != 0 && (t) != (void *)0xdddddddd), \
	assert(*(unsigned int *)(t) != 0xdddddddd)
#else
#define assert_threadpool_valid(t)	assert((t) != 0)
#endif

/*
 * Schedule the function to be run in a thread from the indicated pool.
 * If inkern is non-zero, the thread will be run in the NodeOS context.
 * If no threads are currently available or queueonly is non-zero, the
 * request will be queued.  Returns zero on success, an error code otherwise.
 */
int threadpool_thread_schedule(threadpool_t *tpool, struct ani_chan *chan,
			       int inkern, int queueonly, thread_func_t func,
			       void *farg1, void *farg2);

/*
 * Start a thread from the given pool running the indicated function.
 * If inkern is non-zero, the thread will be run in the NodeOS context.
 * If there are no threads available and the pool is at the thread limit,
 * returns an error.  If the pool is under limit but there are no cached
 * threads available, the call is delayed briefly while a new thread is started.
 * That is, upon return from this call, an actual thread has been commited
 * to run the function.  If out_threadp is non-zero, the call returns a
 * pointer to the new thread.  Returns zero if a thread was successfully
 * started, an error otherwise.
 */
int threadpool_thread_start(threadpool_t *pool, int inkern,
			    thread_func_t func, void *farg1, void *farg2,
			    ani_thread_t **out_thread);

/*
 * Start a thread from the given pool and wait til it terminates or the
 * timeout expires.  If the timeout expires, the thread is killed and a
 * distinguished error code is returned.  If inkern is non-zero, the
 * thread will be run in the NodeOS context.  Returns zero on success,
 * an error code otherwise.
 */
int threadpool_thread_run(threadpool_t *tpool, int inkern,
			  thread_func_t func, void *farg1, void *farg2,
			  an_timespec_t *timo);

int threadpool_thread_setreserve(threadpool_t *tpool, int nthreads, int wait);

#endif /* _THREADPOOL_H_ */
